#! /usr/bin/env python

"""
Module with functions for correcting bad pixels in cubes.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez, V. Christiaens'
__all__ = ['frame_fix_badpix_isolated',
           'cube_fix_badpix_isolated',
           'cube_fix_badpix_annuli',
           'cube_fix_badpix_clump']

import numpy as np
import pyprind
from skimage.draw import circle, ellipse
from scipy.ndimage import median_filter
from astropy.stats import sigma_clipped_stats
from ..stats import sigma_filter
from ..var import dist, frame_center, pp_subplots
from ..stats import clip_array
from ..conf import timing, time_ini


def frame_fix_badpix_isolated(array, bpm_mask=None, sigma_clip=3, num_neig=5,
                              size=5, protect_mask=False, radius=30,
                              verbose=True, debug=False):
    """ Corrects the bad pixels, marked in the bad pixel mask. The bad pixel is
     replaced by the median of the adjacent pixels. This function is very fast
     but works only with isolated (sparse) pixels.

     Parameters
     ----------
     array : array_like
         Input 2d array.
     bpm_mask : array_like, optional
         Input bad pixel map. Zeros frame where the bad pixels have a value of 1.
         If None is provided a bad pixel map will be created per frame using
         sigma clip statistics. In the case of a cube the bad pixels will be
         computed on the mean frame of the stack.
     sigma_clip : int, optional
         In case no bad pixel mask is provided all the pixels above and below
         sigma_clip*STDDEV will be marked as bad.
     num_neig : int, optional
         The side of the square window around each pixel where the sigma clipped
         statistics are calculated (STDDEV and MEDIAN). If the value is equal to
         0 then the statistics are computed in the whole frame.
     size : odd int, optional
         The size the box (size x size) of adjacent pixels for the median filter.
     protect_mask : bool, optional
         If True a circular aperture at the center of the frames will be
         protected from any operation. With this we protect the star and vicinity.
     radius : int, optional
         Radius of the circular aperture (at the center of the frames) for the
         protection mask.
     verbose : bool, optional
         If True additional information will be printed.
     debug : bool, optional
         If debug is True, the bpm_mask and the input array are plotted. If the
         input array is a cube, a long output is to be expected. Better check the
         results with single images.

     Return
     ------
     frame : array_like
         Frame with bad pixels corrected.
     """
    if not array.ndim == 2:
        raise TypeError('Array is not a 2d array or single frame')
    if size % 2 == 0:
        raise TypeError('Size of the median blur kernel must be an odd integer')

    if bpm_mask is not None:
        bpm_mask = bpm_mask.astype('bool')

    if verbose:  start = time_ini()

    if num_neig > 0:
        neigh = True
    else:
        neigh = False

    frame = array.copy()
    cy, cx = frame_center(frame)
    if bpm_mask is None:
        ind = clip_array(frame, sigma_clip, sigma_clip, neighbor=neigh,
                         num_neighbor=num_neig, mad=True)
        bpm_mask = np.zeros_like(frame)
        bpm_mask[ind] = 1
        if protect_mask:
            cir = circle(cy, cx, radius)
            bpm_mask[cir] = 0
        bpm_mask = bpm_mask.astype('bool')
        if debug:
            pp_subplots(frame, bpm_mask, title='Frame / Bad pixel mask')

    smoothed = median_filter(frame, size, mode='mirror')
    frame[np.where(bpm_mask)] = smoothed[np.where(bpm_mask)]
    array_out = frame

    if verbose:
        print("\nDone replacing bad pixels using the median of the neighbors")
        timing(start)
    return array_out


def cube_fix_badpix_isolated(array, bpm_mask=None, sigma_clip=3, num_neig=5, 
                             size=5, protect_mask=False, radius=30, verbose=True,
                             debug=False):
    """ Corrects the bad pixels, marked in the bad pixel mask. The bad pixel is 
    replaced by the median of the adjacent pixels. This function is very fast
    but works only with isolated (sparse) pixels. 
     
    Parameters
    ----------
    array : array_like
        Input 3d array.
    bpm_mask : array_like, optional
        Input bad pixel map. Zeros frame where the bad pixels have a value of 1.
        If None is provided a bad pixel map will be created per frame using 
        sigma clip statistics. In the case of a cube the bad pixels will be 
        computed on the mean frame of the stack.
    sigma_clip : int, optional
        In case no bad pixel mask is provided all the pixels above and below
        sigma_clip*STDDEV will be marked as bad. 
    num_neig : int, optional
        The side of the square window around each pixel where the sigma clipped
        statistics are calculated (STDDEV and MEDIAN). If the value is equal to
        0 then the statistics are computed in the whole frame.
    size : odd int, optional
        The size the box (size x size) of adjacent pixels for the median filter. 
    protect_mask : bool, optional
        If True a circular aperture at the center of the frames will be 
        protected from any operation. With this we protect the star and its
        vicinity.
    radius : int, optional 
        Radius of the circular aperture (at the center of the frames) for the 
        protection mask.
    verbose : bool, optional
        If True additional information will be printed. 
    debug : bool, optional
        If debug is True, the bpm_mask and the input array are plotted. If the
        input array is a cube, a long output is to be expected. Better check the
        results with single images.
    
    Return
    ------
    array_out : array_like
        Cube with bad pixels corrected.
    """
    if not array.ndim == 3:
        raise TypeError('Array is not a 3d array or cube')
    if size % 2 == 0:
        raise TypeError('Size of the median blur kernel must be an odd integer')
    
    if bpm_mask is not None:
        bpm_mask = bpm_mask.astype('bool')
    
    if verbose:  start = time_ini()
    
    if num_neig > 0:
        neigh = True
    else:
        neigh = False
    
    cy, cx = frame_center(array[0])
    cube_out = array.copy()
    n_frames = array.shape[0]

    if bpm_mask is None:
        ind = clip_array(np.mean(array, axis=0), sigma_clip, sigma_clip,
                         neighbor=neigh, num_neighbor=num_neig, mad=True)
        bpm_mask = np.zeros_like(array[0])
        bpm_mask[ind] = 1
        if protect_mask:
            cir = circle(cy, cx, radius)
            bpm_mask[cir] = 0
        bpm_mask = bpm_mask.astype('bool')

    if debug:
        pp_subplots(bpm_mask, title='Bad pixel mask')

    bar = pyprind.ProgBar(n_frames, stream=1, title='Looping through frames')
    for i in range(n_frames):
        frame = cube_out[i]
        smoothed = median_filter(frame, size, mode='mirror')
        frame[np.where(bpm_mask)] = smoothed[np.where(bpm_mask)]
        bar.update()
    array_out = cube_out
    
    if verbose: 
        print("/nDone replacing bad pixels using the median of the neighbors")
        timing(start)
    return array_out


def cube_fix_badpix_annuli(array, cy, cx, fwhm, sig=5., protect_psf=True, 
                           verbose=True, half_res_y=False, min_thr=None, 
                           mid_thr=None, full_output=False):
    """
    Function to correct the bad pixels annulus per annulus (centered on the 
    provided location of the star), in an input frame or cube.
    This function is MUCH FASTER than bp_clump_removal (about 20 times faster);
    hence to be prefered in all cases where there is only one bright source.
    The bad pixel values are replaced by: ann_median + ann_stddev*random_gauss;
    where ann_median is the median of the annulus, ann_stddev is the standard 
    deviation in the annulus, and random_gauss is a random factor picked from a 
    gaussian distribution centered on 0 and with variance 1.

    Parameters
    ----------
    array : 3D or 2D array 
        Input 3d cube or 2d image.
    cy, cx : float or 1D array
        Vector with approximate y and x coordinates of the star for each channel
        (cube_like), or single 2-elements vector (frame_like)
    fwhm: float or 1D array
        Vector containing the full width half maximum of the PSF in pixels, for 
        each channel (cube_like); or single value (frame_like)
    sig: Float scalar, optional
        Number of stddev above or below the median of the pixels in the same 
        annulus, to consider a pixel as bad.
    protect_psf: bool, {True, False}, optional
        Whether to protect a circular region centered on the star (1.8*fwhm 
        radius) from any bpix corr. If False, there is a risk of modifying a 
        centroid peak value if it is too "peaky"; but if True real bad pixels 
        within the core are not corrected.
    verbose: bool, {False, True}, optional
        Whether to print out the number of bad pixels in each frame. 
    half_res_y: bool, {True,False}, optional
        Whether the input data have only half the angular resolution vertically 
        compared to horizontally (e.g. SINFONI data).
        The algorithm will then correct the bad pixels every other row.
    min_thr: {None,float}, optional
        Any pixel whose value is lower than this threshold (expressed in stddev)
        will be automatically considered bad and hence sigma_filtered. If None, 
        it is not used.
    mid_thr: {None, float}, optional
        Pixels whose value is lower than this threshold (expressed in stddev) 
        will have its neighbours checked; if there is at max. 1 neighbour pixel
        whose value is lower than (5+mid_thr)*stddev, then the pixel is 
        considered bad (as it means it is a cold pixel in the middle of 
        significant signal). If None, it is not used.
    full_output: bool, {False,True}, optional
        Whether to return as well the cube of bad pixel maps and the cube of 
        defined annuli.

    Returns:
    --------
    obj_tmp: 2d or 3d array; the bad pixel corrected frame/cube.
    If full_output is set to True, it returns as well:
    bpix_map: 2d or 3d array; the bad pixel map or the cube of bpix maps
    ann_frame_cumul: 2 or 3d array; the cube of defined annuli
    """

    obj_tmp = array.copy()
    ndims = obj_tmp.ndim
    assert ndims == 2 or ndims == 3, "Object is not two or three dimensional.\n"

    #thresholds
    if min_thr is None:
        min_thr = np.amin(obj_tmp)-1
    if mid_thr is None:
        mid_thr = np.amin(obj_tmp)-1

    def bp_removal_2d(obj_tmp, cy, cx, fwhm, sig, protect_psf, verbose):

        n_x = obj_tmp.shape[1]
        n_y = obj_tmp.shape[0]

        # Squash the frame if twice less resolved vertically than horizontally
        if half_res_y:
            if n_y % 2 != 0:
                msg = 'The input frames do not have of an even number of rows. '
                msg2 = 'Hence, you should not use option half_res_y = True'
                raise ValueError(msg+msg2)
            n_y = int(n_y/2)
            cy = int(cy/2.)
            frame = obj_tmp.copy()
            obj_tmp = np.zeros([n_y,n_x])
            for yy in range(n_y):
                obj_tmp[yy] = frame[2*yy]

        #1/ Stddev of background
        _, _, stddev = sigma_clipped_stats(obj_tmp, sigma=2.5)

        #2/ Define each annulus, its median and stddev
        ymax = max(cy,n_y-cy)
        if half_res_y: ymax *= 2.
        xmax = max(cx,n_x-cx)
        rmax = np.sqrt(ymax**2+xmax**2)
        # the annuli definition is optimized for Airy rings
        ann_width = max(1.5,0.61*fwhm) 
        nrad = int(rmax/ann_width)+1
        d_bord_max = max(n_y-cy,cy,n_x-cx,cx)
        if half_res_y:
            d_bord_max = max(2*(n_y-cy),2*cy,n_x-cx,cx)

        big_ell_frame = np.zeros_like(obj_tmp)
        sma_ell_frame = np.zeros_like(obj_tmp)
        ann_frame_cumul = np.zeros_like(obj_tmp)
        n_neig = np.zeros(nrad)
        med_neig = np.zeros(nrad)
        std_neig = np.zeros(nrad)
        neighbours = np.zeros([nrad,n_y*n_x])

        for rr in range(nrad):
            if rr > int(d_bord_max/ann_width):
                # just to merge farthest annuli with very few elements 
                rr_big = nrad  
                rr_sma = int(d_bord_max/ann_width)
            else: 
                rr_big = rr
                rr_sma= rr
            if half_res_y:
                big_ell_idx = ellipse(cy=cy, cx=cx, 
                                      yradius=((rr_big+1)*ann_width)/2., 
                                      xradius=(rr_big+1)*ann_width, 
                                      shape=(n_y,n_x))
                if rr != 0:
                    small_ell_idx = ellipse(cy=cy, cx=cx, 
                                            yradius=(rr_sma*ann_width)/2., 
                                            xradius=rr_sma*ann_width, 
                                            shape=(n_y,n_x))
            else:
                big_ell_idx = circle(cy, cx, radius=(rr_big+1)*ann_width,
                                     shape=(n_y,n_x))
                if rr != 0:
                    small_ell_idx = circle(cy, cx, radius=rr_sma*ann_width, 
                                           shape=(n_y,n_x))
            big_ell_frame[big_ell_idx] = 1
            if rr!=0: sma_ell_frame[small_ell_idx] = 1
            ann_frame = big_ell_frame - sma_ell_frame
            n_neig[rr] = ann_frame[np.where(ann_frame)].shape[0]
            neighbours[rr,:n_neig[rr]] = obj_tmp[np.where(ann_frame)]
            ann_frame_cumul[np.where(ann_frame)] = rr

            # We delete iteratively max and min outliers in each annulus, 
            # so that the annuli median and stddev are not corrupted by bpixs
            neigh = neighbours[rr,:n_neig[rr]]
            n_removal = 0
            n_pix_init = neigh.shape[0]
            while neigh.shape[0] > 10 and n_removal < n_pix_init/10.:
                min_neigh = np.amin(neigh)
                if reject_outliers(neigh, min_neigh, m=5., stddev=stddev,
                                   min_thr=min_thr*stddev,
                                   mid_thr=mid_thr*stddev):
                    min_idx = np.argmin(neigh)
                    neigh = np.delete(neigh,min_idx)
                    n_removal += 1
                else:
                    max_neigh = np.amax(neigh)
                    if reject_outliers(neigh, max_neigh, m=5., stddev=stddev,
                                       min_thr=min_thr*stddev, 
                                       mid_thr=mid_thr*stddev):
                        max_idx = np.argmax(neigh)
                        neigh = np.delete(neigh,max_idx)
                        n_removal += 1
                    else: break
            n_neig[rr] = neigh.shape[0]
            neighbours[rr,:n_neig[rr]] = neigh
            neighbours[rr,n_neig[rr]:] = 0.
            med_neig[rr] = np.median(neigh)
            std_neig[rr] = np.std(neigh)
        
        #3/ Create a tuple-array with coordinates of a circle of radius 1.8*fwhm
        # centered on the provided coordinates of the star
        if protect_psf:
            if half_res_y: 
                circl_new = ellipse(cy, cx, yradius=0.9*fwhm, 
                                    xradius=1.8*fwhm,shape=(n_y,n_x))
            else: 
                circl_new = circle(cy, cx, radius=1.8*fwhm, 
                                   shape=(n_y, n_x))
        else: circl_new = []

        #4/ Loop on all pixels to check bpix
        bpix_map = np.zeros_like(obj_tmp)
        obj_tmp_corr = obj_tmp.copy()
        
        for yy in range(n_y):
            for xx in range(n_x):
                if half_res_y:
                    rad = np.sqrt((2*(cy-yy))**2+(cx-xx)**2)
                else:
                    rad = dist(cy,cx,yy,xx)
                rr = int(rad/ann_width)
                neigh = neighbours[rr,:n_neig[rr]]
                dev = max(stddev,min(std_neig[rr],med_neig[rr]))

                # check min_thr
                if obj_tmp[yy,xx] < min_thr*stddev:
                    bpix_map[yy,xx] = 1
                    # Gaussian noise
                    #obj_tmp_corr[yy,xx] = med_neig[rr] +dev*np.random.randn()
                    # Poisson noise
                    rand_fac= 2.*(np.random.rand()-0.5)
                    obj_tmp_corr[yy,xx] = med_neig[rr] + \
                                          np.sqrt(np.abs(med_neig[rr]))*rand_fac

                # check median +- sig*stddev
                elif (obj_tmp[yy,xx] < med_neig[rr]-sig*dev or 
                      obj_tmp[yy,xx] > med_neig[rr]+sig*dev):
                    bpix_map[yy,xx] = 1  
                    # Gaussian noise
                    #obj_tmp_corr[yy,xx] = med_neig[rr] +dev*np.random.randn()
                    # Poisson noise
                    rand_fac= 2.*(np.random.rand()-0.5)
                    obj_tmp_corr[yy,xx] = med_neig[rr] + \
                                          np.sqrt(np.abs(med_neig[rr]))*rand_fac

                # check mid_thr and neighbours
                else:
                    min_el = max(2.,0.05*n_neig[rr])
                    if (obj_tmp[yy,xx] < mid_thr*stddev and 
                        neigh[neigh<(mid_thr+5.)*stddev].shape[0] < min_el):
                            bpix_map[yy,xx] = 1
                            # Gaussian noise
                            #obj_tmp_corr[yy,xx] = med_neig[rr] + \
                            #                      dev*np.random.randn()
                            # Poisson noise
                            rand_fac = 2.*(np.random.rand()-0.5)
                            obj_tmp_corr[yy,xx] = med_neig[rr] + \
                                                  np.sqrt(np.abs(med_neig[rr]))*rand_fac

        #5/ Count bpix and uncorrect if within the circle
        nbpix_tot = np.sum(bpix_map)
        nbpix_tbc = nbpix_tot - np.sum(bpix_map[circl_new])
        bpix_map[circl_new] = 0
        obj_tmp_corr[circl_new] = obj_tmp[circl_new]
        if verbose:
            print(nbpix_tot, ' bpix in total, and ', nbpix_tbc, ' corrected.')

        # Unsquash all the frames
        if half_res_y:
            frame = obj_tmp_corr.copy()
            frame_bpix = bpix_map.copy()
            n_y = 2*n_y
            obj_tmp_corr = np.zeros([n_y,n_x])
            bpix_map = np.zeros([n_y,n_x])
            ann_frame = ann_frame_cumul.copy()
            ann_frame_cumul = np.zeros([n_y,n_x])
            for yy in range(n_y):
                obj_tmp_corr[yy] = frame[int(yy/2)]
                bpix_map[yy] = frame_bpix[int(yy/2)]
                ann_frame_cumul[yy] = ann_frame[int(yy/2)]

        return obj_tmp_corr, bpix_map, ann_frame_cumul

    if ndims == 2:
        obj_tmp, bpix_map, ann_frame_cumul = bp_removal_2d(obj_tmp, cy, cx, 
                                                           fwhm, sig, 
                                                           protect_psf, verbose)
    if ndims == 3:
        n_z = obj_tmp.shape[0]
        bpix_map = np.zeros_like(obj_tmp)
        ann_frame_cumul = np.zeros_like(obj_tmp)
        if cy.shape[0] != n_z or cx.shape[0] != n_z: 
            raise ValueError('Please provide cy and cx as 1d-arr of size n_z')
        for i in range(n_z):
            if verbose:
                print('************Frame # ', i,' *************')
            res_i = bp_removal_2d(obj_tmp[i], cy[i], cx[i], fwhm[i], sig,
                                  protect_psf, verbose)
            obj_tmp[i], bpix_map[i], ann_frame_cumul[i] = res_i
 
    if full_output:
        return obj_tmp, bpix_map, ann_frame_cumul
    else:
        return obj_tmp


def cube_fix_badpix_clump(array, cy, cx, fwhm, sig=4., protect_psf=True, 
                          verbose=True, half_res_y=False, min_thr=None, 
                          mid_thr=None, max_nit=15, full_output=False):
    """
    Function to correct the bad pixels in obj_tmp. Slow alternative to 
    bp_annuli_removal to correct for bad pixels in either a frame or a spectral
    cube. It should be used instead of bp_annuli_removal only if the observed 
    field is not composed of a single bright object (producing a circularly 
    symmetric PSF). As it is based on an iterative process, it is able to 
    correct clumps of bad pixels. The bad pixel values are sigma filtered 
    (replaced by the median of neighbouring pixel values).


    Parameters
    ----------
    array : 3D or 2D array 
        Input 3d cube or 2d image.
    cy,cx : float or 1D array
        Vector with approximate y and x coordinates of the star for each channel
        (cube_like), or single 2-elements vector (frame_like).
    fwhm: float or 1D array
        Vector containing the full width half maximum of the PSF in pixels, for
        each channel (cube_like); or single value (frame_like)
    sig: float, optional
        Value representing the number of "sigmas" above or below the "median" of
        the neighbouring pixel, to consider a pixel as bad. See details on 
        parameter "m" of function reject_outlier.
    protect_psf: bool, {True, False}, optional
        True if you want to protect a circular region centered on the star 
        (1.8*fwhm radius) from any bpix corr. If False, there is a risk to 
        modify a psf peak value; but if True, real bpix within the core are 
        not corrected.
    verbose: bool, {False,True}, optional
        Whether to print the number of bad pixels and number of iterations 
        required for each frame.
    half_res_y: bool, {True,False}, optional
        Whether the input data has only half the angular resolution vertically 
        compared to horizontally (e.g. the case of SINFONI data); in other words
        there are always 2 rows of pixels with exactly the same values.
        The algorithm will just consider every other row (hence making it
        twice faster), then apply the bad pixel correction on all rows.
    min_thr: {None,float}, optional
        Any pixel whose value is lower than this threshold (expressed in adu)
        will be automatically considered bad and hence sigma_filtered. If None, 
        it is not used (not recommended).
    mid_thr: {None, float}, optional
        Pixels whose value is lower than this threshold (expressed in adu) will
        have its neighbours checked; if there is at max 1 neighbour pixel whose
        value is lower than mid_thr+(5*stddev), then the pixel is considered bad
        (because it means it is a cold pixel in the middle of significant 
        signal). If None, it is not used (not recommended).
    max_nit: float, optional
        Maximum number of iterations on a frame to correct bpix. Typically, it 
        should be set to less than ny/2 or nx/2. This is a mean of precaution in
        case the algorithm gets stuck with 2 neighbouring pixels considered bpix
        alternately on two consecutively iterations hence leading to an infinite
        loop (very very rare case).
    full_output: bool, {False,True}, optional
        Whether to return as well the cube of bad pixel maps and the cube of 
        defined annuli.

    Returns:
    --------
    obj_tmp: 2d or 3d array; the bad pixel corrected frame/cube.
    If full_output is set to True, it returns as well:
    bpix_map: 2d or 3d array; the bad pixel map or the cube of bpix maps
    """

    obj_tmp = array.copy()
    ndims = obj_tmp.ndim
    assert ndims == 2 or ndims == 3, "Object is not two or three dimensional.\n"

    def bp_removal_2d(obj_tmp, cy, cx, fwhm, sig, protect_psf, verbose):    
        n_x = obj_tmp.shape[1]
        n_y = obj_tmp.shape[0]

        if half_res_y:
            if n_y%2 != 0: 
                msg = 'The input frames do not have of an even number of rows. '
                msg2 = 'Hence, you should not use option half_res_y = True'
                raise ValueError(msg+msg2)
            n_y = int(n_y/2)
            frame = obj_tmp.copy()
            obj_tmp = np.zeros([n_y,n_x])
            for yy in range(n_y):
                obj_tmp[yy] = frame[2*yy]
            
        fwhm_round = int(round(fwhm))
        # This should reduce the chance to accidently correct a bright planet:
        if fwhm_round % 2 == 0:
            neighbor_box = max(3,fwhm_round+1) 
        else:
            neighbor_box = max(3,fwhm_round)
        nneig = sum(np.arange(3,neighbor_box+2,2))

        
        #1/ Create a tuple-array with coordinates of a circle of radius 1.8*fwhm
        # centered on the approximate coordinates of the star
        if protect_psf:
            if half_res_y: 
                circl_new = ellipse(int(cy/2), cx, yradius=0.9*fwhm, 
                                    xradius=1.8*fwhm, shape=(n_y, n_x))
            else: circl_new = circle(cy, cx, radius=1.8*fwhm, 
                                     shape=(n_y, n_x))
        else: circl_new = []
        
        #2/ Compute stddev of background
        # empirically 2.5 is best to find stddev of background noise
        _, _, stddev = sigma_clipped_stats(obj_tmp, sigma=2.5) 

        #3/ Create a bad pixel map, by detecting them with find_outliers
        bpix_map = find_outliers(obj_tmp, sig_dist=sig, stddev=stddev, 
                                 neighbor_box=neighbor_box,min_thr=min_thr,
                                 mid_thr=mid_thr)
        nbpix_tot = np.sum(bpix_map)
        nbpix_tbc = nbpix_tot - np.sum(bpix_map[circl_new])
        bpix_map[circl_new] = 0
        bpix_map_cumul = np.zeros_like(bpix_map)
        bpix_map_cumul[:] = bpix_map[:]
        nit = 0

        #4/ Loop over the bpix correction with sigma_filter, until 0 bpix left
        while nbpix_tbc > 0 and nit < max_nit:
            nit = nit+1
            if verbose:
                msg = 'Iteration '+str(nit)+': '+str(nbpix_tot)+\
                      ' bpix in total, '+str(nbpix_tbc)+' to be corrected.'
                print(msg)
            obj_tmp = sigma_filter(obj_tmp, bpix_map, neighbor_box=neighbor_box,
                                   min_neighbors=nneig, verbose=verbose)
            bpix_map = find_outliers(obj_tmp, sig_dist=sig, in_bpix=bpix_map,
                                     stddev=stddev, neighbor_box=neighbor_box,
                                     min_thr=min_thr,mid_thr=mid_thr)
            nbpix_tot = np.sum(bpix_map)
            nbpix_tbc = nbpix_tot - np.sum(bpix_map[circl_new])
            bpix_map[circl_new] = 0
            bpix_map_cumul = bpix_map_cumul+bpix_map

        if verbose:
            print('All bad pixels are corrected.')
            
        if half_res_y:
            frame = obj_tmp.copy()
            frame_bpix = bpix_map_cumul.copy()
            n_y = 2*n_y
            obj_tmp = np.zeros([n_y,n_x])
            bpix_map_cumul = np.zeros([n_y,n_x])
            for yy in range(n_y):
                obj_tmp[yy] = frame[int(yy/2)]
                bpix_map_cumul[yy] = frame_bpix[int(yy/2)]

        return obj_tmp, bpix_map_cumul

    if ndims == 2:
        obj_tmp, bpix_map_cumul = bp_removal_2d(obj_tmp, cy, cx, fwhm, sig, 
                                                protect_psf, verbose)

    if ndims == 3:
        n_z = obj_tmp.shape[0]
        bpix_map_cumul = np.zeros_like(obj_tmp)
        for i in range(n_z):
            if verbose: print('************Frame # ', i,' *************')
            obj_tmp[i], bpix_map_cumul[i] = bp_removal_2d(obj_tmp[i], cy[i], 
                                                          cx[i], fwhm[i], sig, 
                                                          protect_psf, verbose)
    if full_output:
        return obj_tmp, bpix_map_cumul
    else:
        return obj_tmp
    
    
def find_outliers(frame, sig_dist, in_bpix=None, stddev=None,neighbor_box=3,
                  min_thr=None,mid_thr=None):
    """ Provides a bad pixel (or outlier) map for a given frame.

    Parameters
    ----------
    frame: 2d array 
        Input 2d image.
    sig_dist: float
        Threshold used to discriminate good from bad neighbours, in terms of 
        normalized distance to the median value of the set (see reject_outliers)
    in_bpix: 2d array, optional
        Input bpix map (typically known from the previous iteration), to only 
        look for bpix around those locations.
    neighbor_box: int, optional
        The side of the square window around each pixel where the sigma and 
        median are calculated for the bad pixel DETECTION and CORRECTION.
    min_thr: {None,float}, optional
        Any pixel whose value is lower than this threshold (expressed in adu)
        will be automatically considered bad and hence sigma_filtered. If None,
        it is not used.
    mid_thr: {None, float}, optional
        Pixels whose value is lower than this threshold (expressed in adu) will
        have its neighbours checked; if there is at max. 1 neighbour pixel whose
        value is lower than mid_thr+(5*stddev), then the pixel is considered bad
        (because it means it is a cold pixel in the middle of significant 
        signal). If None, it is not used.

    Returns
    -------
    bpix_map : array_like
        Output cube with frames indicating the location of bad pixels"""

    ndims = len(frame.shape)
    assert ndims == 2, "Object is not two dimensional.\n"

    nx = frame.shape[1]
    ny = frame.shape[0]
    bpix_map = np.zeros_like(frame)
    if stddev is None: stddev = np.std(frame)
    half_box = int(np.floor(neighbor_box/2.))
    
    if in_bpix is None:
        for xx in range(nx):
            for yy in range(ny):
                #0/ Determine the box of neighbouring pixels
                # half size of the box at the bottom of the pixel
                hbox_b = min(half_box,yy)        
                # half size of the box at the top of the pixel
                hbox_t = min(half_box,ny-1-yy)   
                # half size of the box to the left of the pixel
                hbox_l = min(half_box,xx)
                # half size of the box to the right of the pixel  
                hbox_r = min(half_box,nx-1-xx)    
                # but in case we are at an edge, we want to extend the box by 
                # one row/column of px in the direction opposite to the edge:
                if yy > ny-1-half_box:
                    hbox_b = hbox_b + (yy-(ny-1-half_box))
                elif yy < half_box:
                    hbox_t = hbox_t+(half_box-yy)
                if xx > nx-1-half_box:
                    hbox_l = hbox_l + (xx-(nx-1-half_box))
                elif xx < half_box:
                    hbox_r = hbox_r+(half_box-xx)

                #1/ list neighbouring pixels, >8 (NOT including pixel itself)
                neighbours = frame[yy-hbox_b:yy+hbox_t+1,
                                   xx-hbox_l:xx+hbox_r+1]
                idx_px = ([[hbox_b],[hbox_l]])
                flat_idx = np.ravel_multi_index(idx_px,(hbox_t+hbox_b+1,
                                                        hbox_r+hbox_l+1))
                neighbours = np.delete(neighbours,flat_idx)

                #2/ Det if central pixel is outlier
                test_result = reject_outliers(neighbours, frame[yy,xx], 
                                              m=sig_dist, stddev=stddev, 
                                              min_thr=min_thr, mid_thr=mid_thr)

                #3/ Assign the value of the test to bpix_map
                bpix_map[yy,xx] = test_result

    else:
        nb = int(np.sum(in_bpix))  # number of bad pixels at previous iteration
        wb = np.where(in_bpix)     # pixels to check
        bool_bpix= np.zeros_like(in_bpix)
        for n in range(nb):
            for yy in [max(0,wb[0][n]-1),wb[0][n],min(ny-1,wb[0][n]+1)]:
                for xx in [max(0,wb[1][n]-1),wb[1][n],min(ny-1,wb[1][n]+1)]:
                    bool_bpix[yy,xx] = 1
        nb = int(np.sum(bool_bpix))# true number of px to check  (including 
                                   # neighbours of bpix from previous iteration)
        wb = np.where(bool_bpix)   # true px to check
        for n in range(nb):
            #0/ Determine the box of neighbouring pixels
            # half size of the box at the bottom of the pixel
            hbox_b = min(half_box,wb[0][n])
            # half size of the box at the top of the pixel     
            hbox_t = min(half_box,ny-1-wb[0][n])
            # half size of the box to the left of the pixel
            hbox_l = min(half_box,wb[1][n])
            # half size of the box to the right of the pixel
            hbox_r = min(half_box,nx-1-wb[1][n])
            # but in case we are at an edge, we want to extend the box by one 
            # row/column of pixels in the direction opposite to the edge:
            if wb[0][n] > ny-1-half_box:
                hbox_b = hbox_b + (wb[0][n]-(ny-1-half_box))
            elif wb[0][n] < half_box:
                hbox_t = hbox_t+(half_box-wb[0][n])
            if wb[1][n] > nx-1-half_box:
                hbox_l = hbox_l + (wb[1][n]-(nx-1-half_box))
            elif wb[1][n] < half_box:
                hbox_r = hbox_r+(half_box-wb[1][n])

            #1/ list neighbouring pixels, > 8, not including the pixel itself
            neighbours = frame[wb[0][n]-hbox_b:wb[0][n]+hbox_t+1,
                               wb[1][n]-hbox_l:wb[1][n]+hbox_r+1]
            c_idx_px = ([[hbox_b],[hbox_l]])
            flat_c_idx = np.ravel_multi_index(c_idx_px,(hbox_t+hbox_b+1,
                                                        hbox_r+hbox_l+1))
            neighbours = np.delete(neighbours,flat_c_idx)

            #2/ test if bpix
            test_result = reject_outliers(neighbours, frame[wb[0][n],wb[1][n]],
                                          m=sig_dist, stddev=stddev, 
                                          min_thr=min_thr, mid_thr=mid_thr)

            #3/ Assign the value of the test to bpix_map
            bpix_map[wb[0][n],wb[1][n]] = test_result


    return bpix_map


def reject_outliers(data, test_value, m=5., stddev=None, debug=False,
                    min_thr=None, mid_thr=None):
    """ FUNCTION TO REJECT OUTLIERS FROM A SET
    Instead of the classic standard deviation criterion (e.g. 5-sigma), the 
    discriminant is determined as follow:
    - for each value in data, an absolute distance to the median of data is
    computed and put in a new array "d" (of same size as data)
    - scaling each element of "d" by the median value of "d" gives the absolute
    distances "s" of each element
    - each "s" is then compared to "m" (parameter): if s < m, we have a good 
    neighbour, otherwise we have an outlier. A specific value test_value is 
    tested as outlier.

    Parameters:
    -----------
    data: array_like
        Input array with respect to which either a test_value or the central a 
        value of data is determined to be an outlier or not
    test_value: float
        Value to be tested as an outlier in the context of the input array data
    m: float, optional
        Criterion used to test if test_value is or pixels of data are outlier(s)
        (similar to the number of "sigma" in std_dev statistics)
    stddev: float, optional (but strongly recommended)
        Global std dev of the non-PSF part of the considered frame. It is needed
        as a reference to know the typical variation of the noise, and hence 
        avoid detecting outliers out of very close pixel values. If the 9 pixels
        of data happen to be very uniform in values at some location, the 
        departure in value of only one pixel could make it appear as a bad 
        pixel. If stddev is not provided, the stddev of data is used (not 
        recommended).
    debug: Bool, {False,True}.
        If True, the different variables involved will be printed out.
    min_thr: {None,float}, optional
        Any pixel whose value is lower than this threshold (expressed in adu)
        will be automatically considered bad and hence sigma_filtered. If None,
        it is not used.
    mid_thr: {None, float}, optional
        Pixels whose value is lower than this threshold (expressed in adu) will
        have its neighbours checked; if there is at max. 1 neighbour pixel whose
        value is lower than mid_thr+(5*stddev), then the pixel is considered bad
        (because it means it is a cold pixel in the middle of significant 
        signal). If None, it is not used.

    Returns:
    --------
    test_result: 0 or 1
        0 if test_value is not an outlier. 1 otherwise. 
    """

    if stddev is None:
        stddev = np.std(data)
    if min_thr is None:
        min_thr = min(np.amin(data), test_value)-1
    if mid_thr is None:
        mid_thr = min(np.amin(data), test_value)-1

    med = np.median(data)
    d = np.abs(data - med)
    mdev = np.median(d)
    if debug:
        print("data = ", data)
        print("median(data)= ", np.median(data))
        print("d = ", d)
        print("mdev = ", mdev)
        print("stddev(box) = ", np.std(data))
        print("stddev(frame) = ", stddev)
        print("max(d) = ", np.max(d))

    n_el = max(2., 0.05*data.shape[0])
    if test_value < min_thr or (test_value < mid_thr and 
                                data[data<(mid_thr+5*stddev)].shape[0] < n_el):
        test_result = 1

    elif max(np.max(d),np.abs(test_value-med)) > stddev:
        mdev = mdev if mdev>stddev else stddev
        s = d/mdev
        if debug:
            print("s =", s)
        test = np.abs((test_value-np.median(data))/mdev)
        if debug:
            print("test =", test)
        else:
            if test < m:
                test_result = 0
            else:
                test_result = 1
    else:
        test_result = 0

    return test_result



#! /usr/bin/env python

"""
Module with contrast curve generation function.
"""

from __future__ import division, print_function

__author__ = 'C. Gomez, O. Absil @ ULg'
__all__ = ['contrast_curve',
           'noise_per_annulus',
           'throughput',
           'aperture_flux']

import numpy as np
import pandas as pd
import photutils
import inspect
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy import stats
from scipy.signal import savgol_filter
from skimage.draw import circle
from matplotlib import pyplot as plt
from .fakecomp import cube_inject_companions, frame_inject_companion, psf_norm
from ..conf import time_ini, timing
from ..conf.utils_conf import sep
from ..var import frame_center, dist


def contrast_curve(cube, angle_list, psf_template, fwhm, pxscale, starphot,
                   algo, sigma=5, nbranch=1, theta=0, inner_rad=1,
                   wedge=(0,360), fc_snr=10.0, student=True, transmission=None,
                   smooth=True, plot=True, dpi=100, imlib='opencv', debug=False,
                   verbose=True, full_output=False, save_plot=None,
                   object_name=None, frame_size=None, fix_y_lim=(),
                   figsize=(8,4), **algo_dict):
    """ Computes the contrast curve at a given SIGMA (``sigma``) level for an
    ADI cube or ADI+IFS cube. The contrast is calculated as
    sigma*noise/throughput. This implementation takes into account the small
    sample statistics correction proposed in Mawet et al. 2014.

    Parameters
    ----------
    cube : array_like
        The input cube, 2d (ADI data) or 3d array (IFS data), without fake
        companions.
    angle_list : array_like
        Vector with the parallactic angles.
    psf_template : array_like
        Frame with the psf template for the fake companion(s).
        PSF must be centered in array. Normalization is done internally.
    fwhm: int or float or 1d array, optional
        The the Full Width Half Maximum in pixels. It can handle a different
        FWHM value for different wavelengths (IFS data).
    pxscale : float
        Plate scale or pixel scale of the instrument.
    starphot : int or float or 1d array
        If int or float it corresponds to the aperture photometry of the
        non-coronagraphic PSF which we use to scale the contrast. If a vector
        is given it must contain the photometry correction for each frame.
    algo : callable or function
        The post-processing algorithm, e.g. vip_hci.pca.pca.
    sigma : int
        Sigma level for contrast calculation.
    nbranch : int optional
        Number of branches on which to inject fakes companions. Each branch
        is tested individually.
    theta : float, optional
        Angle in degrees for rotating the position of the first branch that by
        default is located at zero degrees. Theta counts counterclockwise from
        the positive x axis. When working on a wedge, make sure that theta is
        located inside of it.
    inner_rad : int, optional
        Innermost radial distance to be considered in terms of FWHM.
    wedge : tuple of floats, optional
        Initial and Final angles for using a wedge. For example (-90,90) only
        considers the right side of an image.
    fc_snr: float optional
        Signal to noise ratio of injected fake companions
    student : {True, False}, bool optional
        If True uses Student t correction to inject fake companion.
    transmission : tuple of 2 1d arrays, optional
        If not None, then the tuple contains a vector with the factors to be
        applied to the sensitivity and a vector of the radial distances [px]
        where it is sampled (in this order).
    smooth : {True, False}, bool optional
        If True the radial noise curve is smoothed with a Savitzky-Golay filter
        of order 2.
    plot : {True, False}, bool optional
        Whether to plot the final contrast curve or not. True by default.
    dpi : int optional
        Dots per inch for the plots. 100 by default. 300 for printing quality.
    imlib : {'opencv', 'ndimage-fourier', 'ndimage-interp'}, string optional
        Library or method used for image operations (shifts). Opencv is the
        default for being the fastest.
    debug : {False, True}, bool optional
        Whether to print and plot additional info such as the noise, throughput,
        the contrast curve with different X axis and the delta magnitude instead
        of contrast.
    verbose : {True, False, 0, 1, 2} optional
        If True or 1 the function prints to stdout intermediate info and timing,
        if set to 2 more output will be shown. 
    full_output : {False, True}, bool optional
        If True returns intermediate arrays.
    save_plot: string
        If provided, the contrast curve will be saved to this path.
    object_name: string
        Target name, used in the plot title
    frame_size: int
        Frame size used for generating the contrast curve, used in the plot title
    fix_y_lim: tuple
        If provided, the y axis limits will be fixed, for easier comparison between plots
    **algo_dict
        Any other valid parameter of the post-processing algorithms can be
        passed here.

    Returns
    -------
    datafr : pandas dataframe
        Dataframe containing the sensitivity (Gaussian and Student corrected if
        Student parameter is True), the interpolated throughput, the distance in
        pixels, the noise and the sigma corrected (if Student is True).

    If full_output is True then the function returns: 
        datafr, cube_fc_all, frame_fc_all, frame_nofc and fc_map_all.

    cube_fc_all : array_like
        4d array, with the 3 different pattern cubes with the injected fake
        companions.
    frame_fc_all : array_like
        3d array with the 3 frames of the 3 (patterns) processed cubes with
        companions.
    frame_nofc : array_like
        2d array, PCA processed frame without companions.
    fc_map_all : array_like
        3d array with 3 frames containing the position of the companions in the
        3 patterns.
    """
    if not (cube.ndim == 3 or cube.ndim == 4):
        raise TypeError('The input array is not a 3d or 4d cube')
    if cube.ndim == 3 and (not cube.shape[0] == angle_list.shape[0]):
        raise TypeError('Input parallactic angles vector has wrong length')
    if cube.ndim == 4 and (not cube.shape[1] == angle_list.shape[0]):
        raise TypeError('Input parallactic angles vector has wrong length')
    if cube.ndim == 3 and not psf_template.ndim == 2:
        raise TypeError('Template PSF is not a frame (for ADI case)')
    if cube.ndim == 4 and not psf_template.ndim == 3:
        raise TypeError('Template PSF is not a cube (for ADI+IFS case)')
    if transmission is not None:
        if not isinstance(transmission, tuple) or not len(transmission) == 2:
            raise TypeError('transmission must be a tuple with 2 1d vectors')
    if isinstance(starphot, float) or isinstance(starphot, int):
        pass
    else:
        if not starphot.shape[0] == cube.shape[0]:
            raise TypeError('Correction vector has bad size')
        cube = cube.copy()
        for i in range(cube.shape[0]):
            cube[i] = cube[i] / starphot[i]

    if verbose:
        start_time = time_ini()
        if isinstance(starphot, float) or isinstance(starphot, int):
            msg0 = 'ALGO : {}, FWHM = {}, # BRANCHES = {}, SIGMA = {},'
            msg0 += ' STARPHOT = {}'
            print(msg0.format(algo.__name__, fwhm, nbranch, sigma, starphot))
        else:
            msg0 = 'ALGO : {}, FWHM = {}, # BRANCHES = {}, SIGMA = {}'
            print(msg0.format(algo.__name__, fwhm, nbranch, sigma))
        print(sep)

    # throughput
    verbose_thru = False
    if verbose == 2:
        verbose_thru = True
    res_throug = throughput(cube, angle_list, psf_template, fwhm, pxscale,
                            nbranch=nbranch, theta=theta, inner_rad=inner_rad,
                            wedge=wedge, fc_snr=fc_snr, full_output=True,
                            algo=algo, imlib=imlib, verbose=verbose_thru,
                            **algo_dict)
    vector_radd = res_throug[2]
    if res_throug[0].shape[0] > 1:
        thruput_mean = np.mean(res_throug[0], axis=0)
    else:
        thruput_mean = res_throug[0][0]
    cube_fc_all = res_throug[3]
    frame_fc_all = res_throug[4]
    frame_nofc = res_throug[5]
    fc_map_all = res_throug[6]

    if verbose:
        print('Finished the throughput calculation')
        timing(start_time)

    if thruput_mean[-1] == 0:
        thruput_mean = thruput_mean[:-1]
        vector_radd = vector_radd[:-1]

    # noise measured in the empty PP-frame with better sampling, every px
    # starting from 1*FWHM
    noise_samp, rad_samp = noise_per_annulus(frame_nofc, separation=1, fwhm=fwhm,
                                             init_rad=fwhm, wedge=wedge)
    cutin1 = np.where(rad_samp.astype(int)==vector_radd.astype(int).min())[0][0]
    noise_samp = noise_samp[cutin1:]
    rad_samp = rad_samp[cutin1:]
    cutin2 = np.where(rad_samp.astype(int)==vector_radd.astype(int).max())[0][0]
    noise_samp = noise_samp[:cutin2+1]
    rad_samp = rad_samp[:cutin2+1]

    # interpolating the throughput vector, spline order 2
    f = InterpolatedUnivariateSpline(vector_radd, thruput_mean, k=2)
    thruput_interp = f(rad_samp)

    # interpolating the transmission vector, spline order 1
    if transmission is not None:
        trans = transmission[0]
        radvec_trans = transmission[1]
        f2 = InterpolatedUnivariateSpline(radvec_trans, trans, k=1)
        trans_interp = f2(rad_samp)
        thruput_interp *= trans_interp

    if smooth:
        # smoothing the noise vector using a Savitzky-Golay filter
        win = min(noise_samp.shape[0]-2,int(2*fwhm))
        if win%2 == 0.:
            win += 1
        noise_samp_sm = savgol_filter(noise_samp, polyorder=2, mode='nearest',
                                      window_length=win)
    else:
        noise_samp_sm = noise_samp

    if debug:
        plt.rc("savefig", dpi=dpi)
        plt.figure(figsize=figsize, dpi=dpi)
        plt.plot(vector_radd*pxscale, thruput_mean, '.', label='computed',
                 alpha=0.6)
        plt.plot(rad_samp*pxscale, thruput_interp, ',-', label='interpolated',
                 lw=2, alpha=0.5)
        plt.grid('on', which='both', alpha=0.2, linestyle='solid')
        plt.xlabel('Angular separation [arcsec]')
        plt.ylabel('Throughput')
        plt.legend(loc='best')
        plt.xlim(0, np.max(rad_samp*pxscale))

        plt.figure(figsize=figsize, dpi=dpi)
        plt.plot(rad_samp*pxscale, noise_samp, '.', label='computed', alpha=0.6)
        plt.plot(rad_samp*pxscale, noise_samp_sm, ',-', label='noise smoothed',
                 lw=2, alpha=0.5)
        plt.grid('on', alpha=0.2, linestyle='solid')
        plt.xlabel('Angular separation [arcsec]')
        plt.ylabel('Noise')
        plt.legend(loc='best')
        #plt.yscale('log')
        plt.xlim(0, np.max(rad_samp*pxscale))

    # calculating the contrast
    if isinstance(starphot, float) or isinstance(starphot, int):
        cont_curve_samp = ((sigma * noise_samp_sm)/thruput_interp)/starphot
    else:
        cont_curve_samp = ((sigma * noise_samp_sm)/thruput_interp)
    cont_curve_samp[np.where(cont_curve_samp<0)] = 1
    cont_curve_samp[np.where(cont_curve_samp>1)] = 1

    # calculating the Student corrected contrast
    if student:
        n_res_els = np.floor(rad_samp/fwhm*2*np.pi)
        ss_corr = np.sqrt(1 + 1/(n_res_els-1))
        sigma_corr = stats.t.ppf(stats.norm.cdf(sigma), n_res_els)*ss_corr
        if isinstance(starphot, float) or isinstance(starphot, int):
            cont_curve_samp_corr = ((sigma_corr * noise_samp_sm)/thruput_interp)/starphot
        else:
            cont_curve_samp_corr = ((sigma_corr * noise_samp_sm)/thruput_interp)
        cont_curve_samp_corr[np.where(cont_curve_samp_corr<0)] = 1
        cont_curve_samp_corr[np.where(cont_curve_samp_corr>1)] = 1

    # plotting
    if plot or debug:
        if student:
            label = ['Sensitivity (Gaussian)',
                     'Sensitivity (Student-t correction)']
        else:  label = ['Sensitivity (Gaussian)']

        plt.rc("savefig", dpi=dpi)
        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax1 = fig.add_subplot(111)
        con1, = ax1.plot(rad_samp*pxscale, cont_curve_samp, '-',
                         alpha=0.2, lw=2, color='green')
        con2, = ax1.plot(rad_samp*pxscale, cont_curve_samp, '.',
                         alpha=0.2, color='green')
        if student:
            con3, = ax1.plot(rad_samp*pxscale, cont_curve_samp_corr, '-',
                             alpha=0.4, lw=2, color='blue')
            con4, = ax1.plot(rad_samp*pxscale, cont_curve_samp_corr, '.',
                             alpha=0.4, color='blue')
            lege = [(con1, con2), (con3, con4)]
        else:
            lege = [(con1, con2)]
        plt.legend(lege, label, fancybox=True, fontsize='medium')
        plt.xlabel('Angular separation [arcsec]')
        plt.ylabel(str(sigma)+' sigma contrast')
        plt.grid('on', which='both', alpha=0.2, linestyle='solid')
        ax1.set_yscale('log')
        ax1.set_xlim(0, np.max(rad_samp*pxscale))

        # Give a title to the contrast curve plot
        if object_name is not None and frame_size is not None:
            # Retrieve ncomp and pca_type info to use in title
            ncomp = algo_dict['ncomp']
            if algo_dict['cube_ref'] is None:
                pca_type = 'ADI'
            else:
                pca_type = 'RDI'
            title = pca_type + ' ' + object_name + ' ' + str(ncomp)
            title += 'pc ' + str(frame_size) + '+' + str(inner_rad)
            plt.title(title, fontsize = 14)

        # Option to fix the y-limit
        if len(fix_y_lim) == 2:
            min_y_lim = min(fix_y_lim[0], fix_y_lim[1])
            max_y_lim = max(fix_y_lim[0], fix_y_lim[1])
            ax1.set_ylim(min_y_lim, max_y_lim)

        # Optionally, save the figure to a path
        if save_plot is not None:
            fig.savefig(save_plot, dpi=100)
            
        if debug:
            fig2 = plt.figure(figsize=figsize, dpi=dpi)
            ax3 = fig2.add_subplot(111)
            cc_mags = -2.5*np.log10(cont_curve_samp)
            con4, = ax3.plot(rad_samp*pxscale, cc_mags, '-',
                             alpha=0.2, lw=2, color='green')
            con5, = ax3.plot(rad_samp*pxscale, cc_mags, '.', alpha=0.2,
                             color='green')
            if student:
                cc_mags_corr = -2.5*np.log10(cont_curve_samp_corr)
                con6, = ax3.plot(rad_samp*pxscale, cc_mags_corr, '-',
                                 alpha=0.4, lw=2, color='blue')
                con7, = ax3.plot(rad_samp*pxscale, cc_mags_corr, '.',
                                 alpha=0.4, color='blue')
                lege = [(con4, con5), (con6, con7)]
            else:
                lege = [(con4, con5)]
            plt.legend(lege, label, fancybox=True, fontsize='medium')
            plt.xlabel('Angular separation [arcsec]')
            plt.ylabel('Delta magnitude')
            plt.gca().invert_yaxis()
            plt.grid('on', which='both', alpha=0.2, linestyle='solid')
            ax3.set_xlim(0, np.max(rad_samp*pxscale))
            ax4 = ax3.twiny()
            ax4.set_xlabel('Distance [pixels]')
            ax4.plot(rad_samp, cc_mags, '', alpha=0.)
            ax4.set_xlim(0, np.max(rad_samp))

    if student:
        datafr = pd.DataFrame({'sensitivity (Gauss)': cont_curve_samp,
                               'sensitivity (Student)':cont_curve_samp_corr,
                               'throughput': thruput_interp,
                               'distance': rad_samp, 'noise': noise_samp_sm,
                               'sigma corr':sigma_corr})
    else:
        datafr = pd.DataFrame({'sensitivity (Gauss)': cont_curve_samp,
                               'throughput': thruput_interp,
                               'distance': rad_samp, 'noise': noise_samp_sm})

    if full_output:
        return datafr, cube_fc_all, frame_fc_all, frame_nofc, fc_map_all
    else: 
        return datafr


def throughput(cube, angle_list, psf_template, fwhm, pxscale, algo, nbranch=1,
               theta=0, inner_rad=1, fc_rad_sep=3, wedge=(0,360), fc_snr=10.0,
               full_output=False, imlib='opencv', interpolation='lanczos4',
               verbose=True, **algo_dict):
    """ Measures the throughput for chosen algorithm and input dataset (ADI or
    ADI+IFS). The final throughput is the average of the same procedure measured
    in ``nbranch`` azimutally equidistant branches.

    Parameters
    ---------_
    cube : array_like
        The input cube, 2d (ADI data) or 3d array (IFS data), without fake
        companions.
    angle_list : array_like
        Vector with the parallactic angles.
    psf_template : array_like
        Frame with the psf template for the fake companion(s).
        PSF must be centered in array. Normalization is done internally.
    fwhm: int or float or 1d array, optional
        The the Full Width Half Maximum in pixels. It can handle a different
        FWHM value for different wavelengths (IFS data).
    pxscale : float
        Plate scale in arcsec/px.
    algo : callable or function
        The post-processing algorithm, e.g. vip_hci.pca.pca. Third party Python
        algorithms can be plugged here. They must have the parameters: 'cube',
        'angle_list' and 'verbose'. Optionally a wrapper function can be used.
    nbranch : int optional
        Number of branches on which to inject fakes companions. Each branch
        is tested individually.
    theta : float, optional
        Angle in degrees for rotating the position of the first branch that by
        default is located at zero degrees. Theta counts counterclockwise from
        the positive x axis.
    inner_rad : int, optional
        Innermost radial distance to be considered in terms of FWHM.
    fc_rad_sep : int optional
        Radial separation between the injected companions (in each of the
        patterns) in FWHM. Must be large enough to avoid overlapping. With the
        maximum possible value, a single fake companion will be injected per
        cube and algorithm post-processing (which greatly affects computation
        time).
    wedge : tuple of floats, optional
        Initial and Final angles for using a wedge. For example (-90,90) only
        considers the right side of an image.
    fc_snr: float optional
        Signal to noise ratio of injected fake companions
    full_output : {False, True}, bool optional
        If True returns intermediate arrays.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    verbose : {True, False}, bool optional
        If True prints out timing and information.
    **algo_dict
        Parameters of the post-processing algorithms must be passed here.

    Returns
    -------
    thruput_arr : array_like
        2d array whose rows are the annulus-wise throughput values for each
        branch.
    vector_radd : array_like
        1d array with the distances in FWHM (the positions of the annuli).

    If full_output is True then the function returns: thruput_arr, noise,
    vector_radd, cube_fc_all, frame_fc_all, frame_nofc and fc_map_all.

    noise : array_like
        1d array with the noise per annulus.
    cube_fc_all : array_like
        4d array, with the 3 different pattern cubes with the injected fake
        companions.
    frame_fc_all : array_like
        3d array with the 3 frames of the 3 (patterns) processed cubes with
        companions.
    frame_nofc : array_like
        2d array, PCA processed frame without companions.
    fc_map_all : array_like
        3d array with 3 frames containing the position of the companions in the
        3 patterns.

    """
    array = cube
    parangles = angle_list

    if not (array.ndim == 3 or array.ndim == 4):
        raise TypeError('The input array is not a 3d or 4d cube')
    else:
        if array.ndim == 3:
            if not array.shape[0] == parangles.shape[0]:
                msg = 'Input parallactic angles vector has wrong length'
                raise TypeError(msg)
            if not psf_template.ndim == 2:
                raise TypeError('Template PSF is not a frame or 2d array')
            maxfcsep = int((array.shape[1]/2.)/fwhm)-1
            if not fc_rad_sep >= 3 or not fc_rad_sep <= maxfcsep:
                msg = 'Too large separation between companions in the radial '
                msg += 'patterns. Should lie between 3 and {:}'
                raise ValueError(msg.format(maxfcsep))

        elif array.ndim == 4:
            if not array.shape[1] == parangles.shape[0]:
                msg = 'Input vector or parallactic angles has wrong length'
                raise TypeError(msg)
            if not psf_template.ndim == 3:
                raise TypeError('Template PSF is not a frame, 3d array')
            if 'scale_list' not in algo_dict:
                raise ValueError('Vector of wavelength not found')
            else:
                if not algo_dict['scale_list'].shape[0] == array.shape[0]:
                    raise TypeError('Input wavelength vector has wrong length')
                maxfcsep = int((array.shape[2] / 2.) / fwhm) - 1
                if not fc_rad_sep >= 3 or not fc_rad_sep <= maxfcsep:
                    msg = 'Too large separation between companions in the '
                    msg += 'radial patterns. Should lie between 3 and {:}'
                    raise ValueError(msg.format(maxfcsep))

        # TODO: extend to even-sized psf arrays
        if psf_template.shape[1]%2 == 0:
            raise ValueError("Only odd-sized PSF is accepted")
        if not hasattr(algo, '__call__'):
            raise TypeError('Parameter *algo* must be a callable function')
        if not isinstance(inner_rad, int):
            raise TypeError('inner_rad must be an integer')
        angular_range = wedge[1] - wedge[0]
        if nbranch > 1 and angular_range < 360:
            msg = 'Only a single branch is allowed when working on a wedge'
            raise RuntimeError(msg)

    if verbose:
        start_time = time_ini()
    #***************************************************************************
    # Compute noise in concentric annuli on the "empty frame"
    if 'cube' and 'angle_list' and 'verbose' in inspect.getargspec(algo).args:
        if 'fwhm' in inspect.getargspec(algo).args:
            frame_nofc = algo(cube=array, angle_list=parangles, fwhm=fwhm,
                              verbose=False, **algo_dict)
        else:
            frame_nofc = algo(array, angle_list=parangles, verbose=False,
                              **algo_dict)

    if verbose:
        msg1 = 'Cube without fake companions processed with {:}'
        print(msg1.format(algo.__name__))
        timing(start_time)

    noise, vector_radd = noise_per_annulus(frame_nofc, separation=fwhm,
                                           fwhm=fwhm, wedge=wedge)
    vector_radd = vector_radd[inner_rad-1:]
    noise = noise[inner_rad-1:]
    if verbose:
        print('Measured annulus-wise noise in resulting frame')
        timing(start_time)

    # We crop the PSF and check if PSF has been normalized (so that flux in
    # 1*FWHM aperture = 1) and fix if needed
    new_psf_size = 3 * fwhm
    if new_psf_size % 2 == 0:
        new_psf_size += 1

    if cube.ndim == 3:
        psf_template = psf_norm(psf_template, size=min(new_psf_size,
                                                       psf_template.shape[1]),
                                fwhm=fwhm)

        #***********************************************************************
        # Initialize the fake companions
        angle_branch = angular_range / nbranch
        # signal-to-noise ratio of injected fake companions
        snr_level = fc_snr * np.ones_like(noise)

        thruput_arr = np.zeros((nbranch, noise.shape[0]))
        fc_map_all = np.zeros((nbranch * fc_rad_sep, array.shape[1],
                               array.shape[2]))
        frame_fc_all = fc_map_all.copy()
        cube_fc_all = np.zeros((nbranch * fc_rad_sep, array.shape[0],
                                array.shape[1], array.shape[2]))
        cy, cx = frame_center(array[0])

        # each branch is computed separately
        for br in range(nbranch):
            # each pattern is computed separately. For each one the companions
            # are separated by "fc_rad_sep * fwhm", interleaving the injections
            for irad in range(fc_rad_sep):
                radvec = vector_radd[irad::fc_rad_sep]
                cube_fc = array.copy()
                # filling map with small numbers
                fc_map = np.ones_like(array[0]) * 1e-6
                fcy = []
                fcx = []
                for i in range(radvec.shape[0]):
                    flux = snr_level[irad+i*fc_rad_sep] * noise[irad + i *
                                                                fc_rad_sep]
                    cube_fc = cube_inject_companions(cube_fc, psf_template,
                                                     parangles, flux, pxscale,
                                                     rad_dists=[radvec[i]],
                                                     theta=br*angle_branch +
                                                           theta,
                                                     imlib=imlib, verbose=False,
                                                     interpolation=interpolation)
                    y = cy + radvec[i] * np.sin(np.deg2rad(br * angle_branch +
                                                           theta))
                    x = cx + radvec[i] * np.cos(np.deg2rad(br * angle_branch +
                                                           theta))
                    fc_map = frame_inject_companion(fc_map, psf_template, y, x,
                                                    flux, imlib, interpolation)
                    fcy.append(y)
                    fcx.append(x)

                if verbose:
                    msg2 = 'Fake companions injected in branch {:} '
                    msg2 += '(pattern {:}/{:})'
                    print(msg2.format(br+1, irad+1, fc_rad_sep))
                    timing(start_time)

                #***************************************************************
                if 'cube' and 'angle_list' and 'verbose' in inspect.getargspec(algo).args:
                    if 'fwhm' in inspect.getargspec(algo).args:
                        frame_fc = algo(cube=cube_fc, angle_list=parangles,
                                        fwhm=fwhm, verbose=False, **algo_dict)
                    else:
                        frame_fc = algo(cube=cube_fc, angle_list=parangles,
                                        verbose=False, **algo_dict)
                else:
                    msg = 'Input algorithm must have at least 3 parameters: '
                    msg += 'cube, angle_list and versbose'
                    raise ValueError(msg)

                if verbose:
                    msg3 = 'Cube with fake companions processed with {:}'
                    msg3 += '\nMeasuring its annulus-wise throughput'
                    print(msg3.format(algo.__name__))
                    timing(start_time)

                #***************************************************************
                injected_flux = aperture_flux(fc_map, fcy, fcx, fwhm,
                                              ap_factor=1, mean=False,
                                              verbose=False)
                recovered_flux = aperture_flux((frame_fc - frame_nofc), fcy,
                                               fcx, fwhm, ap_factor=1,
                                               mean=False, verbose=False)
                thruput = recovered_flux / injected_flux
                thruput[np.where(thruput < 0)] = 0

                thruput_arr[br, irad::fc_rad_sep] = thruput
                fc_map_all[br*fc_rad_sep+irad, :, :] = fc_map
                frame_fc_all[br*fc_rad_sep+irad, :, :] = frame_fc
                cube_fc_all[br*fc_rad_sep+irad, :, :, :] = cube_fc

    elif cube.ndim == 4:
        if isinstance(fwhm, (int, float)):
            fwhm = [fwhm for _ in range(array.shape[0])]

        psf_template = psf_norm(psf_template, fwhm=fwhm,
                                size=min(new_psf_size, psf_template.shape[1]))

        # **********************************************************************
        # Initialize the fake companions
        angle_branch = angular_range / nbranch
        # signal-to-noise ratio of injected fake companions
        snr_level = 10.0 * np.ones_like(noise)

        thruput_arr = np.zeros((nbranch, noise.shape[0]))
        fc_map_all = np.zeros((nbranch * fc_rad_sep, array.shape[0],
                               array.shape[2], array.shape[3]))
        frame_fc_all = fc_map_all.copy()
        cube_fc_all = np.zeros(
            (nbranch * fc_rad_sep, array.shape[0], array.shape[1],
             array.shape[2], array.shape[3]))
        cy, cx = frame_center(array[0, 0])

        # each branch is computed separately
        for br in range(nbranch):
            # each pattern is computed separately. For each pattern the
            # companions are separated by "fc_rad_sep * fwhm"
            # radius = vector_radd[irad::fc_rad_sep]
            for irad in range(fc_rad_sep):
                radvec = vector_radd[irad::fc_rad_sep]
                thetavec = range(int(theta), int(theta) + 360,
                                 int(360 / len(radvec)))
                cube_fc = array.copy()
                # filling map with small numbers
                fc_map = np.ones_like(array[:, 0]) * 1e-6
                fcy = []
                fcx = []
                for i in range(radvec.shape[0]):
                    flux = snr_level[irad + i * fc_rad_sep] * noise[irad + i * \
                                                                    fc_rad_sep]
                    cube_fc = cube_inject_companions(cube_fc, psf_template,
                                                     parangles, flux, pxscale,
                                                     rad_dists=[radvec[i]],
                                                     theta=thetavec[i],
                                                     verbose=False)
                    y = cy + radvec[i] * np.sin(np.deg2rad(br * angle_branch +
                                                           thetavec[i]))
                    x = cx + radvec[i] * np.cos(np.deg2rad(br * angle_branch +
                                                           thetavec[i]))
                    fc_map = frame_inject_companion(fc_map, psf_template, y, x,
                                                    flux)
                    fcy.append(y)
                    fcx.append(x)

                if verbose:
                    msg2 = 'Fake companions injected in branch {:} '
                    msg2 += '(pattern {:}/{:})'
                    print(msg2.format(br + 1, irad + 1, fc_rad_sep))
                    timing(start_time)

                # **************************************************************
                if 'cube' and 'angle_list' and 'verbose' in inspect.getargspec(
                        algo).args:
                    if 'fwhm' in inspect.getargspec(algo).args:
                        frame_fc = algo(cube=cube_fc, angle_list=parangles,
                                        fwhm=fwhm, verbose=False, **algo_dict)
                    else:
                        frame_fc = algo(cube=cube_fc, angle_list=parangles,
                                        verbose=False, **algo_dict)

                if verbose:
                    msg3 = 'Cube with fake companions processed with {:}'
                    msg3 += '\nMeasuring its annulus-wise throughput'
                    print(msg3.format(algo.__name__))
                    timing(start_time)

                # **************************************************************
                injected_flux = [aperture_flux(fc_map[i], fcy, fcx, fwhm[i],
                                               ap_factor=1, mean=False,
                                               verbose=False) for i in
                                range(array.shape[0])]
                injected_flux = np.mean(injected_flux, axis=0)
                recovered_flux = aperture_flux((frame_fc - frame_nofc), fcy,
                                               fcx, np.mean(fwhm), ap_factor=1,
                                               mean=False, verbose=False)
                thruput = recovered_flux / injected_flux
                thruput[np.where(thruput < 0)] = 0

                thruput_arr[br, irad::fc_rad_sep] = thruput
                fc_map_all[br * fc_rad_sep + irad, :, :] = fc_map
                frame_fc_all[br * fc_rad_sep + irad, :, :] = frame_fc
                cube_fc_all[br * fc_rad_sep + irad, :, :, :, :] = cube_fc

    if verbose:
        msg = 'Finished measuring the throughput in {:} branches'
        print(msg.format(nbranch))
        timing(start_time)

    if full_output:
        return (thruput_arr, noise, vector_radd, cube_fc_all, frame_fc_all,
                frame_nofc, fc_map_all)
    else:
        return thruput_arr, vector_radd


def noise_per_annulus(array, separation, fwhm, init_rad=None, wedge=(0,360),
                      verbose=False, debug=False):
    """ Measures the noise as the standard deviation of apertures defined in
    each annulus with a given separation.

    Parameters
    ----------
    array : array_like
        Input frame.
    separation : float
        Separation in pixels of the centers of the annuli measured from the
        center of the frame.
    fwhm : float
        FWHM in pixels.
    init_rad : float
        Initial radial distance to be used. If None then the init_rad = FWHM.
    wedge : tuple of floats, optional
        Initial and Final angles for using a wedge. For example (-90,90) only
        considers the right side of an image. Be careful when using small
        wedges, this leads to computing a standard deviation of very small
        samples (<10 values).
    verbose : {False, True}, bool optional
        If True prints information.
    debug : {False, True}, bool optional
        If True plots the positioning of the apertures.

    Returns
    -------
    noise : array_like
        Vector with the noise value per annulus.
    vector_radd : array_like
        Vector with the radial distances values.

    """
    def find_coords(rad, sep, init_angle, fin_angle):
        angular_range = fin_angle-init_angle
        npoints = (np.deg2rad(angular_range)*rad)/sep   #(2*np.pi*rad)/sep
        ang_step = angular_range/npoints   #360/npoints
        x = []
        y = []
        for i in range(int(npoints)):
            newx = rad * np.cos(np.deg2rad(ang_step * i + init_angle))
            newy = rad * np.sin(np.deg2rad(ang_step * i + init_angle))
            x.append(newx)
            y.append(newy)
        return np.array(y), np.array(x)
    #___________________________________________________________________

    if not array.ndim==2:
        raise TypeError('Input array is not a frame or 2d array')
    if not isinstance(wedge, tuple):
        raise TypeError('Wedge must be a tuple with the initial and final angles')

    init_angle, fin_angle = wedge
    centery, centerx = frame_center(array)
    n_annuli = int(np.floor((centery)/separation))

    x = centerx
    y = centery
    noise = []
    vector_radd = []
    if verbose:  print('{} annuli'.format(n_annuli-1))

    if init_rad is None:  init_rad = fwhm

    if debug:
        _, ax = plt.subplots(figsize=(6,6))
        ax.imshow(array, origin='lower', interpolation='nearest',
                  alpha=0.5, cmap='gray')

    for i in range(n_annuli-1):
        y = centery + init_rad + separation*(i)
        rad = dist(centery, centerx, y, x)
        yy, xx = find_coords(rad, fwhm, init_angle, fin_angle)
        yy += centery
        xx += centerx

        apertures = photutils.CircularAperture((xx, yy), fwhm/2.)
        fluxes = photutils.aperture_photometry(array, apertures)
        fluxes = np.array(fluxes['aperture_sum'])

        noise_ann = np.std(fluxes)
        noise.append(noise_ann)
        vector_radd.append(rad)

        if debug:
            for i in range(xx.shape[0]):
                # Circle takes coordinates as (X,Y)
                aper = plt.Circle((xx[i], yy[i]), radius=fwhm/2., color='r',
                              fill=False, alpha=0.8)
                ax.add_patch(aper)
                cent = plt.Circle((xx[i], yy[i]), radius=0.8, color='r',
                              fill=True, alpha=0.5)
                ax.add_patch(cent)

        if verbose:
            print('Radius(px) = {:}, Noise = {:.3f} '.format(rad, noise_ann))

    return np.array(noise), np.array(vector_radd)



def aperture_flux(array, yc, xc, fwhm, ap_factor=1, mean=False, verbose=False):
    """ Returns the sum of pixel values in a circular aperture centered on the
    input coordinates. The radius of the aperture is set as (ap_factor*fwhm)/2.

    Parameters
    ----------
    array : array_like
        Input frame.
    yc, xc : list or 1d arrays
        List of y and x coordinates of sources.
    fwhm : float
        FWHM in pixels.
    ap_factor : int, optional
        Diameter of aperture in terms of the FWHM.

    Returns
    -------
    flux : list of floats
        List of fluxes.

    Note
    ----
    From Photutils documentation, the aperture photometry defines the aperture
    using one of 3 methods:

    'center': A pixel is considered to be entirely in or out of the aperture
              depending on whether its center is in or out of the aperture.
    'subpixel': A pixel is divided into subpixels and the center of each
                subpixel is tested (as above).
    'exact': (default) The exact overlap between the aperture and each pixel is
             calculated.

    """
    n_obj = len(yc)
    flux = np.zeros((n_obj))
    for i, (y, x) in enumerate(zip(yc, xc)):
        if mean:
            ind = circle(y, x,  (ap_factor*fwhm)/2.)
            values = array[ind]
            obj_flux = np.mean(values)
        else:
            aper = photutils.CircularAperture((x, y), (ap_factor*fwhm)/2.)
            obj_flux = photutils.aperture_photometry(array, aper, method='exact')
            obj_flux = np.array(obj_flux['aperture_sum'])
        flux[i] = obj_flux

        if verbose:
            print('Coordinates of object {:} : ({:},{:})'.format(i, y, x))
            print('Object Flux = {:.2f}'.format(flux[i]))

    return flux



#! /usr/bin/env python

"""
Module for image statistics.
"""

from __future__ import division

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['frame_histo_stats']

import numpy as np
from matplotlib import pyplot as plt


def frame_histo_stats(image_array, plot=True):
    """Plots a frame with a colorbar, its histogram and some statistics: mean,
    median, maximum, minimum and standard deviation values.  
    
    Parameters
    ----------
    image_array : array_like
        The input frame.  
    plot : {True, False}, bool optional
        If True plots the frame and the histogram with the values.
        
    Return
    ------
    mean : float
        Mean value of array.
    median : float
        Median value of array.
    std : float
        Standard deviation of array.
    maxim : int or float
        Maximum value.
    minim " int or float
        Minimum value.
        
    """
    vector = image_array.flatten()
    mean = vector.mean()
    median = np.median(vector)
    maxim = vector.max()
    minim = vector.min()
    std = vector.std()
    
    if plot is True:
        fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,5))
        ax0, ax1 = axes.flat
        bins = np.sqrt(vector.shape[0])
        txt = 'Mean = {:.3f}\n'.format(mean) + \
              'Median = {:.3f}\n'.format(median) +\
              'Stddev = {:.3f}\n'.format(std) +\
              'Max = {:.3f}\n'.format(maxim) +\
              'Min = {:.3f}\n\n'.format(minim)
        ax0.hist(vector, bins=bins, label=txt, alpha=0.5, histtype='stepfilled')
        ax0.set_yscale('log')
        ax0.set_title('Histogram')
        ax0.text(0.98, 0.98, txt, transform=ax0.transAxes, fontsize=12,
                 verticalalignment='top', horizontalalignment='right')

        ima = ax1.imshow(image_array, interpolation="nearest", origin ="lower",
                         cmap='CMRmap')
        ax1.set_title('Frame')
        ax1.grid('off')
        fig.colorbar(ima)
        plt.jet()
        plt.show()
        
    return mean, median, std, maxim, minim




    



#! /usr/bin/env python

"""
Module with detection algorithms.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['detection',
           'mask_source_centers',
           'peak_coordinates']

import numpy as np
from matplotlib import pyplot as plt
from scipy.ndimage.filters import correlate
from skimage import feature
from astropy.stats import sigma_clipped_stats
from astropy.stats import gaussian_fwhm_to_sigma, gaussian_sigma_to_fwhm
from astropy.table import Table
from astropy.modeling.models import Gaussian2D
from astropy.modeling.fitting import LevMarLSQFitter
from skimage.feature import peak_local_max
from ..var import (mask_circle, pp_subplots, get_square, frame_center,
                   frame_filter_gaussian2d, fit_2dgaussian)
from ..conf.utils_conf import sep
from .snr import snr_ss
from .frame_analysis import frame_quick_report


# TODO: Add the option of computing and thresholding an S/N map

def detection(array, psf, bkg_sigma=1, mode='lpeaks', matched_filter=False,
              mask=True, snr_thresh=5, plot=True, debug=False,
              full_output=False, verbose=True, save_plot=None,
              object_name = None, frame_size=None, inner_rad=None,
              pca_type = None, ncomp = None, NIRC2angscale = False):
    """ Finds blobs in a 2d array. The algorithm is designed for automatically
    finding planets in post-processed high contrast final frames. Blob can be
    defined as a region of an image in which some properties are constant or
    vary within a prescribed range of values. See <Notes> below to read about
    the algorithm details.

    Parameters
    ----------
    array : array_like, 2d
        Input frame.
    psf : array_like
        Input psf.
    bkg_sigma : float, optional
        The number standard deviations above the clipped median for setting the
        background level.
    mode : {'lpeaks','log','dog'}, optional
        Sets with algorithm to use. Each algorithm yields different results.
    matched_filter : {True, False}, bool optional
        Whether to correlate with the psf of not.
    mask : {True, False}, optional
        Whether to mask the central region (circular aperture of 2*fwhm radius).
    snr_thresh : float, optional
        SNR threshold for deciding whether the blob is a detection or not.
    plot {True, False}, bool optional
        If True plots the frame showing the detected blobs on top.
    debug : {False, True}, bool optional
        Whether to print and plot additional/intermediate results.
    full_output : {False, True}, bool optional
        Whether to output just the coordinates of blobs that fulfill the SNR
        constraint or a table with all the blobs and the peak pixels and SNR.
    verbose : {True,False}, bool optional
        Whether to print to stdout information about found blobs.
    save_plot: string
        If provided, the frames processed by blob detection are saved to that path.
    object_name: string
        Target name, used in the plot title
    frame_size: int
        Frame size of the pca, used in the plot title
    inner_rad: int
        Size of the mask in pca, as a unit of the FWHM, used in the plot title
    pca_type: string
        adi or rdi, used in the title
    ncomp: int
        Number of principal components used to compute the reduced frame, used in the title
    NIRC2angscale: {False, True}
        If True the plot axes are converted to angular scale (arcseconds,
        assuming NIRC2's ~ 0.01 pixel scale)

    Returns
    -------
    yy, xx : array_like
        Two vectors with the y and x coordinates of the centers of the sources
        (potential planets).
    If full_output is True then a table with all the candidates that passed the
    2d Gaussian fit constrains and their SNR is returned. Also the count of
    companions with SNR>5 (those with highest probability of being true
    detections).

    Notes
    -----
    The FWHM of the PSF is measured directly on the provided array. If the
    parameter matched_filter==True then the PSF is used to run a matched filter
    (correlation) which is equivalent to a convolution filter. Filtering the
    image will smooth the noise and maximize detectability of objects with a
    shape similar to the kernel.
    The background level or threshold is found with sigma clipped statistics
    (5 sigma over the median) on the image/correlated image. Then 5 different
    strategies can be used to detect the blobs (potential planets):

    Local maxima + 2d Gaussian fit. The local peaks above the background on the
    (correlated) frame are detected. A maximum filter is used for finding local
    maxima. This operation dilates the original image and merges neighboring
    local maxima closer than the size of the dilation. Locations where the
    original image is equal to the dilated image are returned as local maxima.
    The minimum separation between the peaks is 1*FWHM. A 2d Gaussian fit is
    done on each of the maxima constraining the position on the subimage and the
    sigma of the fit. Finally the blobs are filtered based on its SNR.

    Laplacian of Gaussian + 2d Gaussian fit. It computes the Laplacian of
    Gaussian images with successively increasing standard deviation and stacks
    them up in a cube. Blobs are local maximas in this cube. LOG assumes that
    the blobs are again assumed to be bright on dark. A 2d Gaussian fit is done
    on each of the candidates constraining the position on the subimage and the
    sigma of the fit. Finally the blobs are filtered based on its SNR.

    Difference of Gaussians. This is a faster approximation of LoG approach. In
    this case the image is blurred with increasing standard deviations and the
    difference between two successively blurred images are stacked up in a cube.
    DOG assumes that the blobs are again assumed to be bright on dark. A 2d
    Gaussian fit is done on each of the candidates constraining the position on
    the subimage and the sigma of the fit. Finally the blobs are filtered based
    on its SNR.

    """
    def check_blobs(array_padded, coords_temp, fwhm, debug):
        y_temp = coords_temp[:,0]
        x_temp = coords_temp[:,1]
        coords = []
        # Fitting a 2d gaussian to each local maxima position
        for y,x in zip(y_temp,x_temp):
            subim, suby, subx = get_square(array_padded,
                                           2*int(np.ceil(fwhm)),
                                           y+pad, x+pad, position=True)
            cy, cx = frame_center(subim)

            gauss = Gaussian2D(amplitude=subim.max(),
                               x_mean=cx, y_mean=cy,
                               x_stddev=fwhm*gaussian_fwhm_to_sigma,
                               y_stddev=fwhm*gaussian_fwhm_to_sigma, theta=0)

            sy, sx = np.indices(subim.shape)
            fitter = LevMarLSQFitter()
            fit = fitter(gauss, sx, sy, subim)

            # checking that the amplitude is positive > 0
            # checking whether the x and y centroids of the 2d gaussian fit
            # coincide with the center of the subimage (within 2px error)
            # checking whether the mean of the fwhm in y and x of the fit
            # are close to the FWHM_PSF with a margin of 3px
            fwhm_y = fit.y_stddev.value*gaussian_sigma_to_fwhm
            fwhm_x = fit.x_stddev.value*gaussian_sigma_to_fwhm
            mean_fwhm_fit = np.mean([np.abs(fwhm_x), np.abs(fwhm_y)])
            if fit.amplitude.value>0 \
            and np.allclose(fit.y_mean.value, cy, atol=2) \
            and np.allclose(fit.x_mean.value, cx, atol=2) \
            and np.allclose(mean_fwhm_fit, fwhm, atol=3):
                coords.append((suby+fit.y_mean.value,subx+fit.x_mean.value))

            if debug:
                print('Coordinates (Y,X): {:.3f},{:.3f}'.format(y, x))
                print('fit peak = {:.3f}'.format(fit.amplitude.value))
                #print fit
                msg = 'fwhm_y in px = {:.3f}, fwhm_x in px = {:.3f}'
                print(msg.format(fwhm_y, fwhm_x))
                print('mean fit fwhm = {:.3f}'.format(mean_fwhm_fit))
                pp_subplots(subim, colorb=True)
        return coords

    def print_coords(coords):
        print('Blobs found:', len(coords))
        print(' ycen   xcen')
        print('------ ------')
        for i in range(len(coords[:,0])):
            print('{:.3f} \t {:.3f}'.format(coords[i,0], coords[i,1]))

    def print_abort():
        if verbose:
            print(sep)
            print('No potential sources found')
            print(sep)

    # --------------------------------------------------------------------------

    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array')
    if not psf.ndim == 2 and psf.shape[0] < array.shape[0]:
        raise TypeError('Input psf is not a 2d array or has wrong size')
        
    # Getting the FWHM from the PSF array
    cenpsf = frame_center(psf)
    outdf = fit_2dgaussian(psf, cent=(cenpsf), debug=debug, full_output=True)
    fwhm_x, fwhm_y = outdf.at[0,'fwhm_x'],outdf.at[0,'fwhm_y']
    fwhm = np.mean([fwhm_x, fwhm_y])
    if verbose:
        print('FWHM =', fwhm)
        print()
    if debug:
        print('FWHM_y', fwhm_y)
        print('FWHM_x', fwhm_x)

    # Masking the center, 2*lambda/D is the expected IWA
    if mask: array = mask_circle(array, radius=fwhm)

    # Matched filter
    if matched_filter:
        frame_det = correlate(array, psf)
    else:
        frame_det = array

    # Estimation of background level
    _, median, stddev = sigma_clipped_stats(frame_det, sigma=5, iters=None)
    bkg_level = median + (stddev * bkg_sigma)
    if debug:
        print('Sigma clipped median = {:.3f}'.format(median))
        print('Sigma clipped stddev = {:.3f}'.format(stddev))
        print('Background threshold = {:.3f}'.format(bkg_level))
        print()

    if mode == 'lpeaks' or mode == 'log' or mode == 'dog':
        # Padding the image with zeros to avoid errors at the edges
        pad = 10
        array_padded = np.lib.pad(array, pad, 'constant', constant_values=0)

    if debug and plot and matched_filter:
        print('Input frame after matched filtering:')
        pp_subplots(frame_det, rows=2, colorb=True)

    if mode == 'lpeaks':
        # Finding local peaks (can be done in the correlated frame)
        coords_temp = peak_local_max(frame_det, threshold_abs=bkg_level,
                                     min_distance=int(np.ceil(fwhm)),
                                     num_peaks=20)
        coords = check_blobs(array_padded, coords_temp, fwhm, debug)
        coords = np.array(coords)
        if verbose and coords.shape[0]>0:  print_coords(coords)

    elif mode == 'log':
        sigma = fwhm*gaussian_fwhm_to_sigma
        coords = feature.blob_log(frame_det.astype('float'),
                                  threshold=bkg_level,
                                  min_sigma=sigma-.5, max_sigma=sigma+.5)
        if len(coords)==0:
            print_abort()
            return 0, 0
        coords = coords[:,:2]
        coords = check_blobs(array_padded, coords, fwhm, debug)
        coords = np.array(coords)
        if coords.shape[0]>0 and verbose:  print_coords(coords)

    elif mode == 'dog':
        sigma = fwhm*gaussian_fwhm_to_sigma
        coords = feature.blob_dog(frame_det.astype('float'),
                                  threshold=bkg_level,
                                  min_sigma=sigma-.5, max_sigma=sigma+.5)
        if len(coords)==0:
            print_abort()
            return 0, 0
        coords = coords[:,:2]
        coords = check_blobs(array_padded, coords, fwhm, debug)
        coords = np.array(coords)
        if coords.shape[0]>0 and verbose:  print_coords(coords)

    else:
        msg = 'Wrong mode. Available modes: lpeaks, log, dog.'
        raise TypeError(msg)

    if coords.shape[0]==0:
        print_abort()
        return 0, 0

    yy = coords[:,0]
    xx = coords[:,1]
    yy_final = []
    xx_final = []
    yy_out = []
    xx_out = []
    snr_list = []
    xx -= pad
    yy -= pad

    # Checking SNR for potential sources
    for i in range(yy.shape[0]):
        y = yy[i]
        x = xx[i]
        if verbose:
            print(sep)
            print('X,Y = ({:.1f},{:.1f})'.format(x,y))
        subim = get_square(array, size=15, y=y, x=x)
        snr = snr_ss(array, (x,y), fwhm, False, verbose=False)
        snr_list.append(snr)
        if snr >= snr_thresh:
            if verbose:
                _ = frame_quick_report(array, fwhm, (x,y), verbose=verbose)
            yy_final.append(y)
            xx_final.append(x)
        else:
            yy_out.append(y)
            xx_out.append(x)
            if verbose:  print('S/N constraint NOT fulfilled (S/N = {:.3f})'.format(snr))
            if debug:
                _ = frame_quick_report(array, fwhm, (x,y), verbose=verbose)

    if debug or full_output:
        table = Table([yy.tolist(), xx.tolist(), snr_list],
                      names=('y','x','px_snr'))
        table.sort('px_snr')
    yy_final = np.array(yy_final)
    xx_final = np.array(xx_final)
    yy_out = np.array(yy_out)
    xx_out = np.array(xx_out)

    if plot:
        #print
        #print sep
        #print 'Input frame showing all the detected blobs / potential sources:'
        #print 'In RED circles those that did not pass the SNR and 2dGaussian '
        #print 'fit constraints while in CYAN circles those that passed them.'
        fig, ax = plt.subplots(figsize=(6,6))
        im = ax.imshow(array, origin='lower', interpolation='nearest',
                       cmap='gray', alpha=0.8)

        # Option to plot axes in angular scale
        if NIRC2angscale and frame_size is not None:
            from scipy.ndimage import gaussian_filter
            # Converting axes from pixels to arcseconds
            # Find the middle value in the odd frame sizes
            center_val = int((frame_size / 2.0) + 0.5)
            # Place a tick every 0.5 arcseconds
            half_num_ticks = center_val // 50

            # Calculate the pixel locations at which to put ticks
            ticks = []
            for i in range(half_num_ticks, -half_num_ticks-1, -1):
                # Avoid ticks not showing on the last pixel
                if not center_val - (i) * 50 == frame_size:
                    ticks.append(center_val - (i) * 50)
                else:
                    ticks.append((center_val - (i) * 50) - 1)
                #print xticks
            ax.set_xticks(ticks)
            ax.set_yticks(ticks)

            # Calculate the corresponding distance in arcseconds, measured from the center
            labels = []
            for i in range(half_num_ticks, -half_num_ticks-1, -1):
                labels.append(0.0 - (i) * 0.5)
                #print xlabels
            ax.set_xticklabels(labels)
            ax.set_yticklabels(labels)

            ax.set_xlabel("arcseconds", fontsize=12)
            ax.set_ylabel("arcseconds", fontsize=12)
            plt.tick_params(axis='both', which='major', labelsize=10)

            # Set the title of the plot
            if object_name is not None and inner_rad is not None:
                ax.set_title(pca_type + ' ' + object_name+' '+ str(ncomp) + 'pc ' + str(frame_size)+'+'+str(inner_rad),
                             fontsize=14)
            array_smoothed = gaussian_filter(array, sigma=(2.3, 2.3), order=0)
            plt.imshow(array_smoothed, origin='lower')

        else:
            colorbar_ax = fig.add_axes([0.92, 0.12, 0.03, 0.78])
            fig.colorbar(im, cax=colorbar_ax)
            ax.grid('off')

        for i in range(yy_out.shape[0]):
            y = yy_out[i]
            x = xx_out[i]
            circ = plt.Circle((x, y), radius=fwhm, color='red', fill=False,
                              linewidth=1.5, alpha=0.6)
            ax.text(x, y+1.5*fwhm, (int(x), int(y)), fontsize=10, color='red',
                    family='monospace', ha='center', va='top', weight='bold',
                    alpha=0.6)
            ax.add_patch(circ)
        for i in range(yy_final.shape[0]):
            y = yy_final[i]
            x = xx_final[i]
            circ = plt.Circle((x, y), radius=fwhm, color='cyan', fill=False,
                              linewidth=2)
            ax.text(x, y+1.5*fwhm, (int(x), int(y)), fontsize=10, color='cyan',
                    weight='heavy', family='monospace', ha='center', va='top')
            ax.add_patch(circ)
        # Save the plot if output path is provided
        # Don't show the plot when running pipeline (i.e. when saving figures)
        if save_plot is not None:
            plt.savefig(save_plot, dpi= 100, bbox_inches='tight')
        else:
            plt.show()

    if debug:  print(table)

    if full_output:
        return table, yy_final.shape[0]
    else:
        return yy_final, xx_final


def peak_coordinates(obj_tmp, fwhm, approx_peak=None, search_box=None,
                     channels_peak=False):
    """Find the pixel coordinates of maximum in either a frame or a cube,
    after convolution with gaussian. It first applies a gaussian filter, to
    lower the probability of returning a hot pixel (although it may still
    happen with clumps of hot pixels, hence the need for function
    "approx_stellar_position").

    Parameters
    ----------
    obj_tmp : cube_like or frame_like
        Input 3d cube or image.
    fwhm : float_like
        Input full width half maximum value of the PSF in pixels. This will be
        used as the standard deviation for Gaussian kernel of the Gaussian
        filtering.
    approx_peak: 2 components list or array, opt
        Gives the approximate coordinates of the peak.
    search_box: float or 2 components list or array, opt
        Gives the half-size in pixels of a box in which the peak is searched,
        around approx_peak. If float, it is assumed the same box size is wanted
        in both y and x. Note that this parameter should be provided if
        approx_peak is provided.
    channels_peak: bool, {False, True}, opt
        Whether returns the indices of the peak in each channel in addition to
        the global indices of the peak in the cube. If True, it would hence also
        return two 1d-arrays. (note: only available if the input is a 3d cube)

    Returns
    -------
    zz_max, yy_max, xx_max : integers
        Indices of highest throughput channel

    """

    ndims = len(obj_tmp.shape)
    assert ndims == 2 or ndims == 3, "Array is not two or three dimensional"

    if approx_peak is not None:
        assert len(approx_peak) == 2, "Approx peak is not two dimensional"
        if isinstance(search_box,float) or isinstance(search_box,int):
            sbox_y = search_box
            sbox_x = search_box
        elif len(search_box) == 2:
            sbox_y = search_box[0]
            sbox_x = search_box[1]
        else:
            msg = "The search box does not have the right number of elements"
            raise ValueError(msg)
        if ndims == 3:
            n_z = obj_tmp.shape[0]
            sbox = np.zeros([n_z,2*sbox_y+1,2*sbox_x+1])

    if ndims == 2:
        gauss_filt_tmp = frame_filter_gaussian2d(obj_tmp,
                                            fwhm/gaussian_sigma_to_fwhm)
        if approx_peak is None:
            ind_max = np.unravel_index(gauss_filt_tmp.argmax(),
                                       gauss_filt_tmp.shape)
        else:
            sbox = gauss_filt_tmp[approx_peak[0]-sbox_y:approx_peak[0]+sbox_y+1,
                                  approx_peak[1]-sbox_x:approx_peak[1]+sbox_x+1]
            ind_max_sbox = np.unravel_index(sbox.argmax(), sbox.shape)
            ind_max = (approx_peak[0]-sbox_y+ind_max_sbox[0],
                       approx_peak[1]-sbox_x+ind_max_sbox[1])

        return ind_max

    if ndims == 3:
        n_z = obj_tmp.shape[0]
        gauss_filt_tmp = np.zeros_like(obj_tmp)
        ind_ch_max = np.zeros([n_z,2])

        for zz in range(n_z):
            gauss_filt_tmp[zz] = frame_filter_gaussian2d(obj_tmp[zz],
                                                    fwhm[zz]/gaussian_sigma_to_fwhm)
            if approx_peak is None:
                ind_ch_max[zz] = np.unravel_index(gauss_filt_tmp[zz].argmax(),
                                                  gauss_filt_tmp[zz].shape)
            else:
                sbox[zz] = gauss_filt_tmp[zz, approx_peak[0]-sbox_y:\
                                          approx_peak[0]+sbox_y+1,
                                          approx_peak[1]-sbox_x:\
                                          approx_peak[1]+sbox_x+1]
                ind_max_sbox = np.unravel_index(sbox[zz].argmax(),
                                                sbox[zz].shape)
                ind_ch_max[zz] = (approx_peak[0]-sbox_y+ind_max_sbox[0],
                                  approx_peak[1]-sbox_x+ind_max_sbox[1])

        if approx_peak is None:
            ind_max = np.unravel_index(gauss_filt_tmp.argmax(),
                                       gauss_filt_tmp.shape)
        else:
            ind_max_tmp = np.unravel_index(sbox.argmax(),
                                           sbox.shape)
            ind_max = (ind_max_tmp[0]+approx_peak[0]-sbox_y,
                       ind_max_tmp[1]+approx_peak[1]-sbox_x)

        if channels_peak:
            return ind_max, ind_ch_max
        else:
            return ind_max


def mask_source_centers(array, fwhm, y, x):
    """ Creates a mask of ones with the size of the input frame and zeros at
    the center of the sources (planets) with coordinates x, y.

    Parameters
    ----------
    array : array_like
        Input frame.
    fwhm : float
        Size in pixels of the FWHM.
    y, x : tuples of int
        Coordinates of the center of the sources.

    Returns
    -------
    mask : array_like
        Mask frame.

    """
    if not array.ndim==2:
        raise TypeError('Wrong input array shape.')

    frame = array.copy()
    if not y and x:
        frame = mask_circle(frame, radius=2*fwhm)
        yy, xx = detection(frame, fwhm, plot=False, mode='log')
    else:
        yy = np.array(y); xx = np.array(x)
    mask = np.ones_like(array)
    # center sources become zeros
    mask[yy.astype('int'), xx.astype('int')] = 0
    return mask





#! /usr/bin/env python

"""
PCA algorithm performed on full frame for ADI, RDI or SDI (IFS data).
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['pca',
           'pca_incremental',
           'pca_optimize_snr']

import numpy as np
import pandas as pd
import pyprind
from skimage import draw
from astropy.io import fits
from matplotlib import pyplot as plt
from sklearn.decomposition import IncrementalPCA
from .svd import svd_wrapper
from ..madi.adi_utils import _find_indices, _compute_pa_thresh
from .utils_pca import pca_annulus, scale_cube_for_pca
from ..preproc import (cube_derotate, cube_collapse, check_pa_vector,
                       check_scal_vector)
from ..conf import timing, time_ini, check_enough_memory, get_available_memory
from ..var import frame_center, dist, prepare_matrix, reshape_matrix
from ..stats import descriptive_stats
from .. import phot

import warnings
warnings.filterwarnings("ignore", category=Warning)


def pca(cube, angle_list=None, cube_ref=None, scale_list=None, ncomp=1, ncomp2=1,
        svd_mode='lapack', scaling=None, mask_center_px=None, source_xy=None,
        delta_rot=1, fwhm=4, imlib='opencv', interpolation='lanczos4',
        collapse='median', check_mem=True, full_output=False, verbose=True,
        debug=False):
    """ Algorithm where the reference PSF and the quasi-static speckle pattern 
    are modeled using Principal Component Analysis. Depending on the input
    parameters this PCA function can work in ADI, RDI or SDI (IFS data) mode.
    
    ADI: 
    if neither a reference cube or a scaling vector are provided, the target 
    cube itself is used to learn the PCs and to obtain a low-rank approximation 
    reference PSF (star + speckles).
    
    RDI + ADI: 
    if a reference cube is provided (triggered by *cube_ref* parameter), its 
    PCs are used to project the target frames and to obtain the reference PSF 
    (star + speckles).
    
    SDI (IFS data):
    if a scaling vector is provided (triggered by *scale_list* parameter) and
    the cube is a 3d array, its assumed it contains 1 frame at multiple 
    spectral channels. The frames are re-scaled to match the longest wavelenght.
    Then PCA is applied on this re-scaled cube where the planet will move 
    radially.
    
    SDI (IFS data) + ADI:
    if a scaling vector is provided (triggered by *scale_list* parameter) and
    the cube is a 4d array [# channels, # adi-frames, Y, X], its assumed it 
    contains several multi-spectral ADI frames. A double PCA is performed, first
    on each ADI multi-spectral frame (using ``ncomp`` PCs), then using each ADI
    residual to exploit the rotation (using ``ncomp2`` PCs). If ``ncomp2`` is
    None only one PCA is performed on the ADI multi-spectral frames and then
    the resulting frames are de-rotated and combined.
    
    Several SVD libraries can be used with almost (randsvd stands for randomized 
    SVD) the same result but different computing time.
    
    
    References
    ----------
    KLIP: http://arxiv.org/abs/1207.4197
    pynpoint: http://arxiv.org/abs/1207.6637
    IFS data: http://arxiv.org/abs/1409.6388
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.    
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    scale_list : 
        Scaling factors in case of IFS data. Normally, the scaling factors are
        the central channel wavelength divided by the shortest wavelength in the
        cube. More thorough approaches can be used to get the scaling factors.
    ncomp : int, optional
        How many PCs are used as a lower-dimensional subspace to project the
        target frames. In ADI ncomp is the number of PCs from the target data,
        in RDI ncomp is the number of PCs from the reference data and in IFS 
        ncomp is the number of PCs from the library of spectral channels. 
    ncomp2 : int, optional
        How many PCs are used for IFS+ADI datacubes in the second stage PCA. 
        ncomp2 goes up to the number ADI frames. If None then the second stage
        of the PCA is ignored and the residuals are derotated and combined.
    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy', 'randcupy'}, str
        Switch for the SVD method/library to be used. ``lapack`` uses the LAPACK 
        linear algebra library through Numpy and it is the most conventional way 
        of computing the SVD (deterministic result computed on CPU). ``arpack`` 
        uses the ARPACK Fortran libraries accessible through Scipy (computation
        on CPU). ``eigen`` computes the singular vectors through the 
        eigendecomposition of the covariance M.M' (computation on CPU).
        ``randsvd`` uses the randomized_svd algorithm implemented in Sklearn 
        (computation on CPU). ``cupy`` uses the Cupy library for GPU computation
        of the SVD as in the LAPACK version. ``eigencupy`` offers the same 
        method as with the ``eigen`` option but on GPU (through Cupy). 
        ``randcupy`` is an adaptation of the randomized_svd algorith, where all 
        the computations are done on a GPU. 
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    mask_center_px : None or int
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask. 
    source_xy : tuple of int, optional 
        For ADI PCA, this triggers a frame rejection in the PCA library. 
        source_xy are the coordinates X,Y of the center of the annulus where the
        PA criterion will be used to reject frames from the library. 
    fwhm : float, optional
        Known size of the FHWM in pixels to be used. Default value is 4.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    delta_rot : int, optional
        Factor for increasing the parallactic angle threshold, expressed in FWHM.
        Default is 1 (excludes 1 FHWM on each side of the considered frame).
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    check_mem : {True, False}, bool optional 
        If True, it check that the input cube(s) are smaller than the available
        system memory.
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing. 
    debug : {False, True}, bool optional
        Whether to print debug information or not.
    
    Returns
    -------
    frame : array_like, 2d    
        Median combination of the de-rotated/re-scaled residuals cube.
    
    If full_output is True then it returns: return pcs, recon, residuals_cube, 
    residuals_cube_ and frame. The PCs are not returned when a PA rejection 
    criterion is applied (when *source_xy* is entered). 
    pcs : array_like, 3d
        Cube with the selected principal components.
    recon : array_like, 3d
        Reconstruction of frames using the selected PCs.
    residuals_res : array_like, 3d 
        Cube of residuals.
    residuals_res_der : array_like, 3d
        Cube of residuals after de-rotation/re-scaling.
        
    In the case of IFS+ADI data the it returns: residuals_cube_channels, 
    residuals_cube_channels_ and frame
    residuals_cube_channels : array_like
        Cube with the residuals of the first stage PCA.
    residuals_cube_channels_ : array_like
        Cube with the residuals of the second stage PCA after de-rotation.        
    """
    #***************************************************************************
    # Helping function
    #***************************************************************************
    def subtract_projection(cube, cube_ref, ncomp, scaling, mask_center_px, 
                            debug, svd_mode, verbose, full_output, indices=None,
                            frame=None):
        """ Subtracts the reference PSF after PCA projection. Returns the cube
        of residuals.
        """
        _, y, x = cube.shape
        if indices is not None and frame is not None:
            matrix = prepare_matrix(cube, scaling, mask_center_px, mode='fullfr',
                                    verbose=False)
        else:
            matrix = prepare_matrix(cube, scaling, mask_center_px,
                                    mode='fullfr', verbose=verbose)
              
        if cube_ref is not None:
            ref_lib = prepare_matrix(cube_ref, scaling, mask_center_px,
                                     mode = 'fullfr', verbose=verbose)
        else:
            ref_lib = matrix    
        
        if indices is not None and frame is not None:   # one row (frame) at a time
            ref_lib = ref_lib[indices]
            if ref_lib.shape[0] <= 10:
                msg = 'Too few frames left in the PCA library (<10). '
                msg += 'Try decreasing the parameter delta_rot'
                raise RuntimeError(msg)
            curr_frame = matrix[frame]                     # current frame
    
            V = svd_wrapper(ref_lib, svd_mode, ncomp, False, False)    
            transformed = np.dot(curr_frame, V.T)
            reconstructed = np.dot(transformed.T, V)                        
            residuals = curr_frame - reconstructed 
            if full_output:
                return ref_lib.shape[0], residuals, reconstructed
            else:
                return ref_lib.shape[0], residuals
        else:   # the whole matrix
            V = svd_wrapper(ref_lib, svd_mode, ncomp, debug, verbose)  
                  
            if verbose: timing(start_time)
            transformed = np.dot(V, matrix.T)
            reconstructed = np.dot(transformed.T, V)
            residuals = matrix - reconstructed
            residuals_res = reshape_matrix(residuals,y,x)
            if full_output:
                return residuals_res, reconstructed, V
            else:
                return residuals_res
    
    #***************************************************************************
    # Validation of input parameters
    #***************************************************************************
    if not cube.ndim>2:
        raise TypeError('Input array is not a 3d or 4d array')
    if angle_list is not None:
        if (cube.ndim==3 and not (cube.shape[0] == angle_list.shape[0])) or  \
           (cube.ndim==4 and not (cube.shape[1] == angle_list.shape[0])):
            msg = "Angle list vector has wrong length. It must equal the number"
            msg += " frames in the cube"
            raise TypeError(msg)
    if source_xy is not None and delta_rot is None or fwhm is None:
        msg = 'Delta_rot or fwhm parameters missing. They are needed for the ' 
        msg += 'PA-based rejection of frames from the library'  
        raise TypeError(msg)
    if cube_ref is not None:
        if not cube_ref.ndim==3:
            raise TypeError('Input reference array is not a cube or 3d array')
        if not cube_ref.shape[1]==cube.shape[1]:
            msg = 'Frames in reference cube and target cube have different size'
            raise TypeError(msg)
        if scale_list is not None:
            raise RuntimeError('RDI + SDI (IFS) is not a valid mode')
        n, y, x = cube_ref.shape
    else:
        if scale_list is not None:
            if cube.ndim==3:
                z, y_in, x_in = cube.shape
            if cube.ndim==4:
                z, n, y_in, x_in = cube.shape
        else:
            n, y, x = cube.shape
    if angle_list is None and scale_list is None:
        msg = 'Either the angles list of scale factors list must be provided'
        raise ValueError(msg)
    if scale_list is not None:
        if np.array(scale_list).ndim>1:
            raise TypeError('Wrong scaling factors list. Must be a vector')
        if not scale_list.shape[0]==cube.shape[0]:
            raise TypeError('Scaling factors vector has wrong length')
    
    if verbose:
        start_time = time_ini()
    
    if check_mem:
        input_bytes = cube.nbytes
        if cube_ref is not None:
            input_bytes += cube_ref.nbytes
        if not check_enough_memory(input_bytes, 1.5, False):
            msgerr = 'Input cubes are larger than available system memory. '
            msgerr += 'Set check_mem=False to override this memory check or '
            msgerr += 'use the incremental PCA (for ADI)'
            raise RuntimeError(msgerr)
        
    if angle_list is not None:
        angle_list = check_pa_vector(angle_list)

    #***************************************************************************
    # scale_list triggers SDI(IFS)
    #***************************************************************************
    if scale_list is not None:
        if ncomp > z:
            ncomp = min(ncomp, z)
            msg = 'Number of PCs too high (max PCs={}), using instead {:} PCs.'
            print(msg.format(z, ncomp))
        scale_list = check_scal_vector(scale_list)
        
        #***********************************************************************
        # RDI (IFS): case of 3d cube with multiple spectral channels
        #***********************************************************************
        if cube.ndim==3:
            if verbose:  
                print('{:} spectral channels in IFS cube'.format(z))
            # cube has been re-scaled to have the planets moving radially
            cube, _, y, x, _, _ = scale_cube_for_pca(cube, scale_list)
            residuals_result = subtract_projection(cube, None, ncomp, scaling, 
                                                   mask_center_px,debug,svd_mode, 
                                                   verbose, full_output)
            if full_output:
                residuals_cube = residuals_result[0]
                reconstructed = residuals_result[1] 
                V = residuals_result[2]
                pcs = reshape_matrix(V, y, x)
                recon = reshape_matrix(reconstructed, y, x)
                residuals_cube_,frame,_,_,_,_ = scale_cube_for_pca(residuals_cube, 
                                                        scale_list,
                                                        full_output=full_output,
                                                        inverse=True, y_in=y_in, 
                                                        x_in=x_in)
            else:
                residuals_cube = residuals_result
                frame = scale_cube_for_pca(residuals_cube, scale_list,
                                           full_output=full_output,
                                           inverse=True, y_in=y_in, x_in=x_in)
            if verbose:
                print('Done re-scaling and combining')
                timing(start_time)
        
        #***********************************************************************
        # SDI (IFS) + ADI: cube with multiple spectral channels + rotation
        # shape of cube: [# channels, # adi-frames, Y, X]
        #***********************************************************************
        elif cube.ndim == 4 and angle_list is not None:
            if verbose:  
                print('{:} spectral channels in IFS cube'.format(z))
                print('{:} ADI frames in all channels'.format(n))
            residuals_cube_channels = np.zeros((n, y_in, x_in))
            
            bar = pyprind.ProgBar(n, stream=1, 
                                  title='Looping through ADI frames')
            for i in range(n):
                cube_res, _, y, x, _, _ = scale_cube_for_pca(cube[:,i,:,:], 
                                                             scale_list)
                residuals_result = subtract_projection(cube_res, None, ncomp,
                                                       scaling, mask_center_px, 
                                                       debug, svd_mode, False, 
                                                       full_output)
                if full_output:
                    residuals_cube = residuals_result[0]
                    _,frame,_,_,_,_ = scale_cube_for_pca(residuals_cube, 
                                                         scale_list,
                                                         full_output=full_output,
                                                         inverse=True, y_in=y_in, 
                                                         x_in=x_in)
                else:
                    residuals_cube = residuals_result
                    frame = scale_cube_for_pca(residuals_cube, scale_list,
                                               full_output=full_output,
                                               inverse=True,y_in=y_in,x_in=x_in)
                
                residuals_cube_channels[i] = frame
                bar.update()
            
            # de-rotation of the PCA processed channels, ADI fashion
            if ncomp2 is None:
                residuals_cube_channels_ = cube_derotate(residuals_cube_channels,
                                                    angle_list, imlib=imlib,
                                                    interpolation=interpolation)
                frame = cube_collapse(residuals_cube_channels_, mode=collapse)
                if verbose:
                    print('De-rotating and combining')
                    timing(start_time)
            else:
                if ncomp2 > n:
                    ncomp2 = min(ncomp2, n)
                    msg = 'Number of PCs too high (max PCs={}), using instead {:} PCs.'
                    print(msg.format(n, ncomp2))
                res_ifs_adi = subtract_projection(residuals_cube_channels, None,
                                                  ncomp2, scaling, mask_center_px,
                                                  debug, svd_mode, False,
                                                  full_output)
                residuals_cube_channels_ = cube_derotate(res_ifs_adi,
                                                         angle_list,
                                                         imlib=imlib,
                                                         interpolation=interpolation)
                frame = cube_collapse(residuals_cube_channels_, mode=collapse)
                if verbose:
                    msg = 'Done PCA per ADI multi-spectral frame, de-rotating '
                    msg += 'and combining'
                    print(msg)
                    timing(start_time)

    #***************************************************************************
    # cube_ref triggers RDI+ADI
    #***************************************************************************
    elif cube_ref is not None:
        if ncomp > n:
            ncomp = min(ncomp,n)
            msg = 'Number of PCs too high (max PCs={}), using instead {:} PCs.'
            print(msg.format(n, ncomp))
        residuals_result = subtract_projection(cube, cube_ref, ncomp, scaling, 
                                               mask_center_px, debug, svd_mode, 
                                               verbose, full_output)
        if full_output:
            residuals_cube = residuals_result[0]
            reconstructed = residuals_result[1]
            V = residuals_result[2]
            pcs = reshape_matrix(V, y, x)
            recon = reshape_matrix(reconstructed, y, x)
        else:
            residuals_cube = residuals_result
        residuals_cube_ = cube_derotate(residuals_cube, angle_list, imlib=imlib,
                                        interpolation=interpolation)
        frame = cube_collapse(residuals_cube_, mode=collapse)

        if verbose:
            print('Done de-rotating and combining')
            timing(start_time)
            
    #***************************************************************************
    # normal ADI PCA 
    #***************************************************************************     
    else:
        if ncomp > n:
            ncomp = min(ncomp,n)
            msg = 'Number of PCs too high (max PCs={}), using instead {:} PCs.'
            print(msg.format(n, ncomp))
        
        if source_xy is None:
            residuals_result = subtract_projection(cube, None, ncomp, scaling, 
                                                   mask_center_px, debug, 
                                                   svd_mode,verbose,full_output)
            if full_output: 
                residuals_cube = residuals_result[0]
                reconstructed = residuals_result[1] 
                V = residuals_result[2]
                pcs = reshape_matrix(V, y, x)
                recon = reshape_matrix(reconstructed, y, x)
            else:
                residuals_cube = residuals_result
        else:
            nfrslib = []
            residuals_cube = np.zeros_like(cube)
            recon_cube = np.zeros_like(cube)
            yc, xc = frame_center(cube[0], False)
            x1, y1 = source_xy
            ann_center = dist(yc, xc, y1, x1)
            pa_thr = _compute_pa_thresh(ann_center, fwhm, delta_rot)
            mid_range = np.abs(np.amax(angle_list) - np.amin(angle_list))/2
            if pa_thr >= mid_range - mid_range * 0.1:
                new_pa_th = float(mid_range - mid_range * 0.1)
                if verbose:
                    msg = 'PA threshold {:.2f} is too big, will be set to {:.2f}'
                    print(msg.format(pa_thr, new_pa_th))
                pa_thr = new_pa_th     
            
            for frame in range(n):
                if ann_center > fwhm*3: # TODO: 3 optimal value? new parameter?
                    ind = _find_indices(angle_list, frame, pa_thr,
                                        truncate=True)
                else:
                    ind = _find_indices(angle_list, frame, pa_thr)
                
                res_result = subtract_projection(cube, None, ncomp, scaling, 
                                                 mask_center_px, debug, 
                                                 svd_mode, verbose, full_output,
                                                 ind, frame)
                if full_output:
                    nfrslib.append(res_result[0])
                    residual_frame = res_result[1]
                    recon_frame = res_result[2]
                    residuals_cube[frame] = residual_frame.reshape(cube[0].shape) 
                    recon_cube[frame] = recon_frame.reshape(cube[0].shape) 
                else:
                    nfrslib.append(res_result[0])
                    residual_frame = res_result[1]
                    residuals_cube[frame] = residual_frame.reshape(cube[0].shape) 
            
            # number of frames in library printed for each annular quadrant
            if verbose:
                descriptive_stats(nfrslib, verbose=verbose, label='Size LIB: ')

        residuals_cube_ = cube_derotate(residuals_cube, angle_list, imlib=imlib,
                                        interpolation=interpolation)
        frame = cube_collapse(residuals_cube_, mode=collapse)
        if verbose:
            print('Done de-rotating and combining')
            timing(start_time)
        
    if full_output and cube.ndim<4:
        if source_xy is not None:
            return recon_cube, residuals_cube, residuals_cube_, frame
        else:
            return pcs, recon, residuals_cube, residuals_cube_, frame
    elif full_output and cube.ndim==4:
        return residuals_cube_channels, residuals_cube_channels_, frame
    else:
        return frame
    
    
    
def pca_optimize_snr(cube, angle_list, source_xy, fwhm, cube_ref=None,
                     mode='fullfr', annulus_width=20, range_pcs=None,
                     svd_mode='lapack', scaling=None, mask_center_px=None, 
                     fmerit='px', min_snr=0, imlib='opencv',
                     interpolation='lanczos4', collapse='median', verbose=True,
                     full_output=False, debug=False, plot=True, save_plot=None,
                     plot_title=None):
    """ Optimizes the number of principal components by doing a simple grid 
    search measuring the SNR for a given position in the frame (ADI, RDI). 
    The metric used could be the given pixel's SNR, the maximum SNR in a FWHM 
    circular aperture centered on the given coordinates or the mean SNR in the 
    same circular aperture. They yield slightly different results.
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    source_xy : tuple of floats
        X and Y coordinates of the pixel where the source is located and whose
        SNR is going to be maximized.
    fwhm : float 
        Size of the PSF's FWHM in pixels.
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging. 
    mode : {'fullfr', 'annular'}, optional
        Mode for PCA processing (full-frame or just in an annulus). There is a
        catch: the optimal number of PCs in full-frame may not coincide with the
        one in annular mode. This is due to the fact that the annulus matrix is
        smaller (less noisy, probably not containing the central star) and also
        its intrinsic rank (smaller that in the full frame case).
    annulus_width : float, optional
        Width in pixels of the annulus in the case of the "annular" mode. 
    range_pcs : tuple, optional
        The interval of PCs to be tried. If None then the algorithm will find
        a clever way to sample from 1 to 200 PCs. If a range is entered (as 
        (PC_INI, PC_MAX)) a sequential grid will be evaluated between PC_INI
        and PC_MAX with step of 1. If a range is entered (as 
        (PC_INI, PC_MAX, STEP)) a grid will be evaluated between PC_INI and 
        PC_MAX with the given STEP.          
    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy', 'randcupy'}, str
        Switch for the SVD method/library to be used. ``lapack`` uses the LAPACK 
        linear algebra library through Numpy and it is the most conventional way 
        of computing the SVD (deterministic result computed on CPU). ``arpack`` 
        uses the ARPACK Fortran libraries accessible through Scipy (computation
        on CPU). ``eigen`` computes the singular vectors through the 
        eigendecomposition of the covariance M.M' (computation on CPU).
        ``randsvd`` uses the randomized_svd algorithm implemented in Sklearn 
        (computation on CPU). ``cupy`` uses the Cupy library for GPU computation
        of the SVD as in the LAPACK version. ``eigencupy`` offers the same 
        method as with the ``eigen`` option but on GPU (through Cupy). 
        ``randcupy`` is an adaptation of the randomized_svd algorith, where all 
        the computations are done on a GPU. 
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    mask_center_px : None or int, optional
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask.  
    fmerit : {'px', 'max', 'mean'}
        The function of merit to be maximized. 'px' is *source_xy* pixel's SNR, 
        'max' the maximum SNR in a FWHM circular aperture centered on 
        *source_xy* and 'mean' is the mean SNR in the same circular aperture.  
    min_snr : float
        Value for the minimum acceptable SNR. Setting this value higher will 
        reduce the steps.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing.
    full_output : {False, True} bool optional
        If True it returns the optimal number of PCs, the final PCA frame for 
        the optimal PCs and a cube with all the final frames for each number 
        of PC that was tried.
    debug : {False, True}, bool optional
        Whether to print debug information or not.
    plot : {True, False}, optional
        Whether to plot the SNR and flux as functions of PCs and final PCA 
        frame or not.
    save_plot: string
        If provided, the pc optimization plot will be saved to that path.
    plot_title: string
        If provided, the plot is titled

    Returns
    -------
    opt_npc : int
        Optimal number of PCs for given source.
    If full_output is True, the final processed frame, and a cube with all the
    PCA frames are returned along with the optimal number of PCs.
    """    
    def truncate_svd_get_finframe(matrix, angle_list, ncomp, V):
        """ Projection, subtraction, derotation plus combination in one frame.
        Only for full-frame"""
        transformed = np.dot(V[:ncomp], matrix.T)
        reconstructed = np.dot(transformed.T, V[:ncomp])
        residuals = matrix - reconstructed
        frsize = int(np.sqrt(matrix.shape[1]))                                  # only for square frames
        residuals_res = reshape_matrix(residuals, frsize, frsize)
        residuals_res_der = cube_derotate(residuals_res, angle_list, imlib=imlib,
                                          interpolation=interpolation)
        frame = cube_collapse(residuals_res_der, mode=collapse)
        return frame

    def truncate_svd_get_finframe_ann(matrix, indices, angle_list, ncomp, V):
        """ Projection, subtraction, derotation plus combination in one frame.
        Only for annular mode"""
        transformed = np.dot(V[:ncomp], matrix.T)
        reconstructed = np.dot(transformed.T, V[:ncomp])
        residuals_ann = matrix - reconstructed
        residuals_res = np.zeros_like(cube)
        residuals_res[:,indices[0],indices[1]] = residuals_ann
        residuals_res_der = cube_derotate(residuals_res, angle_list, imlib=imlib,
                                          interpolation=interpolation)
        frame = cube_collapse(residuals_res_der, mode=collapse)
        return frame

    def get_snr(matrix, angle_list, y, x, mode, V, fwhm, ncomp, fmerit,
                full_output):
        if mode=='fullfr':
            frame = truncate_svd_get_finframe(matrix, angle_list, ncomp, V)
        elif mode=='annular':
            frame = truncate_svd_get_finframe_ann(matrix, annind, angle_list,
                                                  ncomp, V)
        else:
            raise RuntimeError('Wrong mode. Choose either full or annular')
        
        if fmerit=='max':
            yy, xx = draw.circle(y, x, fwhm/2.)
            res = [phot.snr_ss(frame, (x_,y_), fwhm, plot=False, verbose=False, 
                               full_output=True) for y_, x_ in zip(yy, xx)]
            snr_pixels = np.array(res)[:,-1]
            fluxes = np.array(res)[:,2]
            argm = np.argmax(snr_pixels)
            if full_output:
                # integrated fluxes for the max snr
                return np.max(snr_pixels), fluxes[argm], frame
            else:
                return np.max(snr_pixels), fluxes[argm]
        elif fmerit=='px':
            res = phot.snr_ss(frame, (x,y), fwhm, plot=False, verbose=False,
                              full_output=True)
            snrpx = res[-1]
            fluxpx = np.array(res)[2]
            if full_output:
                # integrated fluxes for the given px
                return snrpx, fluxpx, frame
            else:
                return snrpx, fluxpx
        elif fmerit=='mean':
            yy, xx = draw.circle(y, x, fwhm/2.)
            res = [phot.snr_ss(frame, (x_,y_), fwhm, plot=False, verbose=False, 
                               full_output=True) for y_, x_ in zip(yy, xx)]  
            snr_pixels = np.array(res)[:,-1]
            fluxes = np.array(res)[:,2]
            if full_output:
                # mean of the integrated fluxes (shifting the aperture)
                return np.mean(snr_pixels), np.mean(fluxes), frame
            else:                         
                return np.mean(snr_pixels), np.mean(fluxes)
    
    def grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, step, inti, intf, 
             debug, full_output, truncate=True):
        nsteps = 0
        snrlist = []
        pclist = []
        fluxlist = []
        if full_output:  frlist = []
        counter = 0
        if debug:  
            print('Step current grid:', step)
            print('PCs | SNR')
        for pc in range(inti, intf+1, step):
            if full_output:
                snr, flux, frame = get_snr(matrix, angle_list, y, x, mode, V,
                                           fwhm, pc, fmerit, full_output)
            else:
                snr, flux = get_snr(matrix, angle_list, y, x, mode, V, fwhm, pc,
                                    fmerit, full_output)
            if np.isnan(snr):  snr=0
            if nsteps>1 and snr<snrlist[-1]:  counter += 1
            snrlist.append(snr)
            pclist.append(pc)
            fluxlist.append(flux)
            if full_output:  frlist.append(frame)
            nsteps += 1
            if truncate and nsteps>2 and snr<min_snr:  
                if debug:  print('SNR too small')
                break
            if debug:  print('{} {:.3f}'.format(pc, snr))
            if truncate and counter==5:  break 
        argm = np.argmax(snrlist)
        
        if len(pclist)==2: pclist.append(pclist[-1]+1)
    
        if debug:
            print('Finished current stage')
            try:
                pclist[argm+1]
                print('Interval for next grid: ', pclist[argm-1], 'to',
                      pclist[argm+1])
            except:
                print('The optimal SNR seems to be outside of the given PC range')
            print()
        
        if argm==0:  argm = 1 
        if full_output:
            return argm, pclist, snrlist, fluxlist, frlist
        else: 
            return argm, pclist, snrlist, fluxlist
    
    #---------------------------------------------------------------------------
    if not cube.ndim==3:
        raise TypeError('Input array is not a cube or 3d array')
    
    if verbose: start_time = time_ini()
    n = cube.shape[0]
    x, y = source_xy 
    
    if range_pcs is not None:
        if len(range_pcs)==2:
            pcmin, pcmax = range_pcs
            pcmax = min(pcmax, n)
            step = 1
        elif len(range_pcs)==3:
            pcmin, pcmax, step = range_pcs
            pcmax = min(pcmax, n)
        else:
            msg = 'Range_pcs tuple must be entered as (PC_INI, PC_MAX, STEP) '
            msg += 'or (PC_INI, PC_MAX)'
            raise TypeError(msg)
    else:
        pcmin = 1
        pcmax = 200
        pcmax = min(pcmax, n)
    
    # Getting `pcmax` principal components a single time
    if mode=='fullfr':
        matrix = prepare_matrix(cube, scaling, mask_center_px, verbose=False)
        if cube_ref is not None:
            ref_lib = prepare_matrix(cube_ref, scaling, mask_center_px,
                                     verbose=False)
        else:
            ref_lib = matrix

    elif mode=='annular':
        y_cent, x_cent = frame_center(cube[0])
        ann_radius = dist(y_cent, x_cent, y, x)
        matrix, annind = prepare_matrix(cube, scaling, None, mode='annular',
                                        annulus_radius=ann_radius,
                                        annulus_width=annulus_width,
                                        verbose=False)
        if cube_ref is not None:
            ref_lib, _ = prepare_matrix(cube_ref, scaling, mask_center_px,
                                     mode='annular', annulus_radius=ann_radius,
                                     annulus_width=annulus_width, verbose=False)
        else:
            ref_lib = matrix

    else:
        raise RuntimeError('Wrong mode. Choose either fullfr or annular')

    V = svd_wrapper(ref_lib, svd_mode, pcmax, False, verbose)

    # sequential grid
    if range_pcs is not None:
        grid1 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, step, 
                     pcmin, pcmax, debug, full_output, False)
        if full_output:
            argm, pclist, snrlist, fluxlist, frlist = grid1
        else:
            argm, pclist, snrlist, fluxlist = grid1
        
        opt_npc = pclist[argm]    
        if verbose:
            print('Number of steps', len(pclist))
            msg = 'Optimal number of PCs = {}, for SNR={}'
            print(msg.format(opt_npc, snrlist[argm]))
            print()
            timing(start_time)
        
        if full_output: 
            cubeout = np.array((frlist))

        # Plot of SNR as function of PCs  
        if plot:    
            plt.figure(figsize=(8,4))
            ax1 = plt.subplot(211)     
            ax1.plot(pclist, snrlist, '-', alpha=0.5)
            ax1.plot(pclist, snrlist, 'o', alpha=0.5, color='blue')
            ax1.set_xlim(np.array(pclist).min(), np.array(pclist).max())
            ax1.set_ylim(0, np.array(snrlist).max()+1)
            ax1.set_ylabel('SNR')
            ax1.minorticks_on()
            ax1.grid('on', 'major', linestyle='solid', alpha=0.4)
            
            ax2 = plt.subplot(212)
            ax2.plot(pclist, fluxlist, '-', alpha=0.5, color='green')
            ax2.plot(pclist, fluxlist, 'o', alpha=0.5, color='green')
            ax2.set_xlim(np.array(pclist).min(), np.array(pclist).max())
            ax2.set_ylim(0, np.array(fluxlist).max()+1)
            ax2.set_xlabel('Principal components')
            ax2.set_ylabel('Flux in FWHM ap. [ADUs]')
            ax2.minorticks_on()
            ax2.grid('on', 'major', linestyle='solid', alpha=0.4)
            print()
            
    # automatic "clever" grid
    else:
        grid1 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, 
                     max(int(pcmax*0.1),1), pcmin, pcmax, debug, full_output)
        if full_output:
            argm, pclist, snrlist, fluxlist, frlist1 = grid1
        else:
            argm, pclist, snrlist, fluxlist = grid1
        
        grid2 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, 
                     max(int(pcmax*0.05),1), pclist[argm-1], pclist[argm+1], debug, 
                     full_output)
        if full_output:
            argm2, pclist2, snrlist2, fluxlist2, frlist2 = grid2
        else:
            argm2, pclist2, snrlist2, fluxlist2 = grid2
        
        grid3 = grid(matrix, angle_list, y, x, mode, V, fwhm, fmerit, 1, 
                     pclist2[argm2-1], pclist2[argm2+1], debug, full_output, 
                     False)
        if full_output:
            _, pclist3, snrlist3, fluxlist3, frlist3 = grid3
        else:
            _, pclist3, snrlist3, fluxlist3 = grid3
        
        argm = np.argmax(snrlist3)
        opt_npc = pclist3[argm]    
        dfr = pd.DataFrame(np.array((pclist+pclist2+pclist3, 
                                     snrlist+snrlist2+snrlist3,
                                     fluxlist+fluxlist2+fluxlist3)).T)  
        dfrs = dfr.sort_values(0)
        dfrsrd = dfrs.drop_duplicates()
        ind = np.array(dfrsrd.index)    
        
        if verbose:
            print('Number of evaluated steps', ind.shape[0])
            msg = 'Optimal number of PCs = {}, for SNR={}'
            print(msg.format(opt_npc, snrlist3[argm]), '\n')
            timing(start_time)
        
        if full_output: 
            cubefrs = np.array((frlist1+frlist2+frlist3))
            cubeout = cubefrs[ind]
    
        # Plot of SNR as function of PCs  
        if plot:   
            alpha = 0.4
            lw = 2
            plt.figure(figsize=(6,4))   
            ax1 = plt.subplot(211)  
            ax1.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,1]), '-', 
                     alpha=alpha, color='blue', lw=lw)
            ax1.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,1]), 'o',  
                     alpha=alpha/2, color='blue')
            ax1.set_xlim(np.array(dfrsrd.loc[:,0]).min(), np.array(dfrsrd.loc[:,0]).max())
            ax1.set_ylim(0, np.array(dfrsrd.loc[:,1]).max()+1)
            #ax1.set_xlabel('')
            ax1.set_ylabel('S/N')
            ax1.minorticks_on()
            ax1.grid('on', 'major', linestyle='solid', alpha=0.2)
            if plot_title is not None:
                ax1.set_title('Optimal pc: ' + str(opt_npc) + ' for ' + plot_title)
            
            ax2 = plt.subplot(212)
            ax2.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,2]), '-', 
                     alpha=alpha, color='green', lw=lw)
            ax2.plot(np.array(dfrsrd.loc[:,0]), np.array(dfrsrd.loc[:,2]), 'o', 
                     alpha=alpha/2, color='green')
            ax2.set_xlim(np.array(pclist).min(), np.array(pclist).max())
            #ax2.set_ylim(0, np.array(fluxlist).max()+1)
            ax2.set_xlabel('Principal components')
            ax2.set_ylabel('Flux in FWHM aperture')
            ax2.minorticks_on()
            ax2.set_yscale('log')
            ax2.grid('on', 'major', linestyle='solid', alpha=0.2)
            #plt.savefig('figure.pdf', dpi=300, bbox_inches='tight')
            print()
    
    # Optionally, save the contrast curve
    if save_plot != None:
        plt.savefig(save_plot, dpi=100, bbox_inches='tight')

    if mode == 'fullfr':
        finalfr = pca(cube, angle_list, cube_ref=cube_ref, ncomp=opt_npc,
                      svd_mode=svd_mode, mask_center_px=mask_center_px,
                      scaling=scaling, imlib=imlib, interpolation=interpolation,
                      collapse=collapse, verbose=False)
    elif mode == 'annular':
        finalfr = pca_annulus(cube, angle_list, ncomp=opt_npc,
                              annulus_width=annulus_width, r_guess=ann_radius,
                              cube_ref=cube_ref, svd_mode=svd_mode,
                              scaling=scaling, collapse=collapse, imlib=imlib,
                              interpolation=interpolation)

    _ = phot.frame_quick_report(finalfr, fwhm, (x,y), verbose=verbose)
    
    if full_output:
        return opt_npc, finalfr, cubeout
    else:
        return opt_npc


def pca_incremental(cubepath, angle_list=None, n=0, batch_size=None, 
                    batch_ratio=0.1, ncomp=10, imlib='opencv',
                    interpolation='lanczos4', collapse='median',
                    verbose=True, full_output=False):
    """ Computes the full-frame PCA-ADI algorithm in batches, for processing 
    fits files larger than the available system memory. It uses the incremental 
    PCA algorithm from scikit-learn. 
    
    Parameters
    ----------
    cubepath : str
        String with the path to the fits file to be opened in memmap mode.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame. If None the parallactic
        angles are obtained from the same fits file (extension). 
    n : int optional
        The index of the HDULIST contaning the data/cube.
    batch_size : int optional
        The number of frames in each batch. If None the size of the batch is 
        computed wrt the available memory in the system.
    batch_ratio : float
        If batch_size is None, batch_ratio indicates the % of the available 
        memory that should be used by every batch.
    ncomp : int, optional
        How many PCs are used as a lower-dimensional subspace to project the
        target frames.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing. 
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
        
    Returns
    -------
    If full_output is True the algorithm returns the incremental PCA model of
    scikit-learn, the PCs reshaped into images, the median of the derotated 
    residuals for each batch, and the final frame. If full_output is False then
    the final frame is returned.
    
    """
    if verbose:  start = time_ini()
    if not isinstance(cubepath, str):
        msgerr = 'Cubepath must be a string with the full path of your fits file'
        raise TypeError(msgerr)
      
    fitsfilename = cubepath
    hdulist = fits.open(fitsfilename, memmap=True)
    if not hdulist[n].data.ndim>2:
        raise TypeError('Input array is not a 3d or 4d array')
    
    n_frames = hdulist[n].data.shape[0]
    y = hdulist[n].data.shape[1]
    x = hdulist[n].data.shape[2]
    if angle_list is None:
        try:  
            angle_list = hdulist[n+1].data
        except:  
            raise RuntimeError('Parallactic angles were not provided')
    if not n_frames==angle_list.shape[0]:
        msg ='Angle list vector has wrong length. It must equal the number of \
        frames in the cube.'
        raise TypeError(msg)
    
    ipca = IncrementalPCA(n_components=ncomp)
    
    if batch_size is None:
        aval_mem = get_available_memory(verbose)
        total_size = hdulist[n].data.nbytes
        batch_size = int(n_frames/(total_size/(batch_ratio*aval_mem)))
    
    if verbose:
        msg1 = "Cube with {:} frames ({:.3f} GB)"
        print(msg1.format(n_frames, hdulist[n].data.nbytes/1e9))
        msg2 = "Batch size set to {:} frames ({:.3f} GB)"
        print(msg2.format(batch_size, hdulist[n].data[:batch_size].nbytes/1e9),
              '\n')
                
    res = n_frames%batch_size
    for i in range(0, int(n_frames/batch_size)):
        intini = i*batch_size
        intfin = (i+1)*batch_size
        batch = hdulist[n].data[intini:intfin]
        msg = 'Processing batch [{},{}] with shape {}'
        if verbose:
            print(msg.format(intini, intfin, batch.shape))
            print('Batch size in memory = {:.3f} MB'.format(batch.nbytes/1e6))
        matrix = prepare_matrix(batch, verbose=False)
        ipca.partial_fit(matrix)
    if res>0:
        batch = hdulist[n].data[intfin:]
        msg = 'Processing batch [{},{}] with shape {}'
        if verbose:
            print(msg.format(intfin, n_frames, batch.shape))
            print('Batch size in memory = {:.3f} MB'.format(batch.nbytes/1e6))
        matrix = prepare_matrix(batch, verbose=False)
        ipca.partial_fit(matrix)
    
    if verbose:  timing(start)
    
    V = ipca.components_
    mean = ipca.mean_.reshape(batch.shape[1], batch.shape[2])
    
    if verbose:
        print('\nReconstructing and obtaining residuals')
    medians = []
    for i in range(0, int(n_frames/batch_size)):
        intini = i*batch_size
        intfin = (i+1)*batch_size
        batch = hdulist[n].data[intini:intfin]
        batch = batch - mean 
        matrix = prepare_matrix(batch, verbose=False)
        reconst = np.dot(np.dot(matrix, V.T), V)
        resid = matrix - reconst
        resid_der = cube_derotate(resid.reshape(batch.shape[0], 
                                                batch.shape[1],
                                                batch.shape[2]), 
                                  angle_list[intini:intfin], imlib=imlib,
                                  interpolation=interpolation)
        medians.append(cube_collapse(resid_der, mode=collapse))
    if res>0:
        batch = hdulist[n].data[intfin:]
        batch = batch - mean
        matrix = prepare_matrix(batch, verbose=False)
        reconst = np.dot(np.dot(matrix, V.T), V)
        resid = matrix - reconst
        resid_der = cube_derotate(resid.reshape(batch.shape[0], 
                                                batch.shape[1],
                                                batch.shape[2]), 
                                  angle_list[intfin:], imlib=imlib,
                                  interpolation=interpolation)
        medians.append(cube_collapse(resid_der, mode=collapse))
    del(matrix)
    del(batch)

    medians = np.array(medians)
    frame = np.median(medians, axis=0)
    
    if verbose:  timing(start)

    if full_output:
        pcs = reshape_matrix(V, y, x)
        return ipca, pcs, medians, frame
    else:
        return frame
    
    
